@@ -25,6 +25,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "quakedef.h"
 #include "sound.h"
 #include "sys.h"
+#include "vid.h"
 #include "view.h"
 
 entity_t r_worldentity;
@@ -81,11 +82,22 @@ cvar_t r_mirroralpha = { "r_mirroralpha", "1" };
 cvar_t r_wateralpha = { "r_wateralpha", "1" };
 cvar_t r_dynamic = { "r_dynamic", "1" };
 cvar_t r_novis = { "r_novis", "0" };
+cvar_t r_waterwarp = { "r_waterwarp", "1" };
 
 cvar_t r_fullbright = {
-    .name     = "r_fullbright",
-    .string   = "0",
-    .readonly = true
+    .name = "r_fullbright",
+    .string = "0",
+    .flags = CVAR_DEVELOPER
+};
+cvar_t gl_keeptjunctions = {
+    .name = "gl_keeptjunctions",
+    .string = "1",
+    .flags = CVAR_OBSOLETE
+};
+cvar_t gl_reporttjunctions = {
+    .name = "gl_reporttjunctions",
+    .string = "0",
+    .flags = CVAR_OBSOLETE
 };
 
 cvar_t gl_finish = { "gl_finish", "0" };
@@ -98,15 +110,17 @@ cvar_t gl_polyblend = { "gl_polyblend", "1" };
 cvar_t gl_flashblend = { "gl_flashblend", "1" };
 cvar_t gl_playermip = { "gl_playermip", "0" };
 cvar_t gl_nocolors = { "gl_nocolors", "0" };
-cvar_t gl_keeptjunctions = { "gl_keeptjunctions", "0" };
-cvar_t gl_reporttjunctions = { "gl_reporttjunctions", "0" };
 cvar_t gl_doubleeyes = { "gl_doubleeyes", "1" };
 
+cvar_t _gl_lightmap_sort = { "_gl_lightmap_sort", "0" };
+cvar_t _gl_sky_mtex = { "_gl_sky_mtex", "1" };
+cvar_t _gl_allowgammafallback = { "_gl_allowgammafallback", "1" };
+
 /*
 =================
 R_CullBox
 
-Returns true if the box is completely outside the frustom
+Returns true if the box is completely outside the frustum
 =================
 */
 qboolean
@@ -321,7 +335,6 @@ GL_DrawAliasFrame(aliashdr_t *paliashdr, int posenum)
     }
 }
 
-
 /*
 =============
 GL_DrawAliasShadow
@@ -420,7 +433,6 @@ R_SetupAliasFrame(int frame, aliashdr_t *paliashdr)
 /*
 =================
 R_DrawAliasModel
-
 =================
 */
 static void
@@ -464,7 +476,7 @@ R_DrawAliasModel(entity_t *e)
 
 	    if (add > 0) {
 		ambientlight += add;
-		//ZOID models should be affected by dlights as well
+		/* ZOID: models should be affected by dlights as well */
 		shadelight += add;
 	    }
 	}
@@ -727,6 +739,21 @@ R_DrawViewModel(void)
     glDepthRange(gldepthmin, gldepthmax);
 }
 
+/*
+ * GL_DrawBlendPoly
+ * - Render a polygon covering the whole screen
+ * - Used for full-screen color blending and approximated gamma correction
+ */
+static void
+GL_DrawBlendPoly(void)
+{
+    glBegin(GL_QUADS);
+    glVertex3f(10, 100, 100);
+    glVertex3f(10, -100, 100);
+    glVertex3f(10, -100, -100);
+    glVertex3f(10, 100, -100);
+    glEnd();
+}
 
 /*
 ============
@@ -736,36 +763,46 @@ R_PolyBlend
 static void
 R_PolyBlend(void)
 {
-    if (!gl_polyblend.value)
-	return;
-    if (!v_blend[3])
-	return;
+    float gamma = 1.0;
 
-    GL_DisableMultitexture();
+    if (!VID_IsFullScreen() || (!VID_SetGammaRamp &&
+				_gl_allowgammafallback.value)) {
+	gamma = v_gamma.value * v_gamma.value;
+	if (gamma < 0.25)
+	    gamma = 0.25;
+	else if (gamma > 1.0)
+	    gamma = 1.0;
+    }
 
-    glDisable(GL_ALPHA_TEST);
-    glEnable(GL_BLEND);
-    glDisable(GL_DEPTH_TEST);
-    glDisable(GL_TEXTURE_2D);
+    if ((gl_polyblend.value && v_blend[3]) || gamma < 1.0) {
+	GL_DisableMultitexture();
 
-    glLoadIdentity();
+	glDisable(GL_ALPHA_TEST);
+	glEnable(GL_BLEND);
+	glDisable(GL_DEPTH_TEST);
+	glDisable(GL_TEXTURE_2D);
 
-    glRotatef(-90, 1, 0, 0);	// put Z going up
-    glRotatef(90, 0, 0, 1);	// put Z going up
+	glLoadIdentity();
+	glRotatef(-90, 1, 0, 0);	// put Z going up
+	glRotatef(90, 0, 0, 1);		// put Z going up
 
-    glColor4fv(v_blend);
+	if (gl_polyblend.value && v_blend[3]) {
+	    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+	    glColor4fv(v_blend);
+	    GL_DrawBlendPoly();
+	}
+	if (gamma < 1.0) {
+	    glBlendFunc(GL_DST_COLOR, GL_ONE_MINUS_SRC_ALPHA);
+	    glColor4f(1, 1, 1, gamma);
+	    GL_DrawBlendPoly();
+	}
 
-    glBegin(GL_QUADS);
+	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 
-    glVertex3f(10, 100, 100);
-    glVertex3f(10, -100, 100);
-    glVertex3f(10, -100, -100);
-    glVertex3f(10, 100, -100);
-    glEnd();
-
-    glDisable(GL_BLEND);
-    glEnable(GL_TEXTURE_2D);
-    glEnable(GL_ALPHA_TEST);
+	glDisable(GL_BLEND);
+	glEnable(GL_TEXTURE_2D);
+	glEnable(GL_ALPHA_TEST);
+    }
 }
 
 
@@ -1012,7 +1049,7 @@ R_Clear(void)
 	gldepthmax = 0.5;
 	glDepthFunc(GL_LEQUAL);
     } else if (gl_ztrick.value) {
-	static int trickframe;
+	static unsigned int trickframe = 0;
 
 	if (gl_clear.value)
 	    glClear(GL_COLOR_BUFFER_BIT);
@@ -1127,8 +1164,10 @@ R_RenderView(void)
     if (!r_worldentity.model || !cl.worldmodel)
 	Sys_Error("%s: NULL worldmodel", __func__);
 
-    if (r_speeds.value) {
+    if (gl_finish.value || r_speeds.value)
 	glFinish();
+
+    if (r_speeds.value) {
 	time1 = Sys_DoubleTime();
 	c_brush_polys = 0;
 	c_alias_polys = 0;
@@ -1137,9 +1176,6 @@ R_RenderView(void)
 
     mirror = false;
 
-    if (gl_finish.value)
-	glFinish();
-
     R_Clear();
 
     // render normal view
