@@ -889,10 +889,11 @@ SCR_UpdateScreen(void)
 	Sbar_FinaleOverlay();
 	SCR_CheckDrawCenterString();
     } else {
-	if (crosshair.value)
+	if (crosshair.value) {
+	    //Draw_Crosshair();
 	    Draw_Character(scr_vrect.x + scr_vrect.width / 2,
 			   scr_vrect.y + scr_vrect.height / 2, '+');
-
+	}
 	SCR_DrawRam();
 	SCR_DrawNet();
 	SCR_DrawTurtle();
