@@ -278,6 +278,10 @@ Host_WriteConfiguration(void)
 	Key_WriteBindings(f);
 	Cvar_WriteVariables(f);
 
+	/* Save the mlook state (rarely used as an actual key binding) */
+	if (in_mlook.state & 1)
+	    fprintf(f, "+mlook\n");
+
 	fclose(f);
     }
 }
@@ -424,7 +428,7 @@ Host_ShutdownServer(qboolean crash)
     int i;
     int count;
     sizebuf_t buf;
-    char message[4];
+    byte message[4];
     double start;
 
     if (!sv.active)
@@ -899,9 +903,15 @@ Host_Init(quakeparms_t *parms)
 	if (!host_colormap)
 	    Sys_Error("Couldn't load gfx/colormap.lmp");
 
-	//#ifndef _WIN32                        // on non win32, mouse comes before video for security reasons
-	//IN_Init ();
-	//#endif
+#if 0
+#ifndef _WIN32
+	/*
+	 * On non win32, mouse comes before video for security reasons
+	 * FIXME - huh?
+	 */
+	IN_Init ();
+#endif
+#endif
 
 	VID_Init(host_basepal);
 
@@ -910,8 +920,11 @@ Host_Init(quakeparms_t *parms)
 	R_Init();
 
 #ifndef	_WIN32
-	// on Win32, sound initialization has to come before video initialization, so we
-	// can put up a popup if the sound hardware is in use
+	/*
+	 * On Win32, sound initialization has to come before video
+	 * initialization, so we can put up a popup if the sound hardware is
+	 * in use
+	 */
 	S_Init();
 #else
 
@@ -931,11 +944,12 @@ Host_Init(quakeparms_t *parms)
 	//#endif
     }
 
-    Cbuf_InsertText("exec quake.rc\n");
-
     Hunk_AllocName(0, "-HOST_HUNKLEVEL-");
     host_hunklevel = Hunk_LowMark();
 
+    Cbuf_InsertText("exec quake.rc\n");
+    Cbuf_Execute();
+
     host_initialized = true;
 
     Sys_Printf("========Quake Initialized=========\n");
