@@ -130,7 +130,7 @@ cvar_t r_lockfrustum = { "r_lockfrustum", "0" };
 cvar_t r_fullbright = {
     .name = "r_fullbright",
     .string = "0",
-    .readonly = true
+    .flags = CVAR_DEVELOPER
 };
 
 static cvar_t r_timegraph = { "r_timegraph", "0" };
