@@ -969,7 +969,7 @@ Tell all the clients that the server is changing levels
 void
 SV_SendReconnect(void)
 {
-    char data[128];
+    byte data[128];
     sizebuf_t msg;
 
     msg.data = data;
@@ -1147,6 +1147,7 @@ SV_SpawnServer(char *server)
     sv.state = ss_active;
 
 // run two frames to allow everything to settle
+    /* FIXME - don't do this when loading a saved game */
     host_frametime = 0.1;
     SV_Physics();
     SV_Physics();
