@@ -93,32 +93,34 @@ viddef_t vid;			// global video state
 #define VID_MODE_NONE			(-1)
 #define VID_MODE_WINDOWED		0
 #define NUM_WINDOWED_MODES		5
-#define VID_MODE_FULLSCREEN_DEFAULT	(VID_MODE_WINDOWED+NUM_WINDOWED_MODES)
+
+/* so this can be "stringified" -> (VID_MODE_WINDOWED+NUM_WINDOWED_MODES) */
+#define VID_MODE_FULLSCREEN_DEFAULT	5
 
 static cvar_t vid_mode = {
-    "vid_mode",
-    stringify(VID_MODE_WINDOWED),
-    false
+    .name = "vid_mode",
+    .string = stringify(VID_MODE_WINDOWED),
+    .archive = false
 };
 static cvar_t _vid_default_mode = {
-    "_vid_default_mode",
-    stringify(VID_MODE_WINDOWED),
-    true
+    .name = "_vid_default_mode",
+    .string = stringify(VID_MODE_WINDOWED),
+    .archive = true
 };
 static cvar_t _vid_default_mode_win = {
-    "_vid_default_mode_win",
-    stringify(VID_MODE_FULLSCREEN_DEFAULT),
-    true
+    .name = "_vid_default_mode_win",
+    .string = stringify(VID_MODE_FULLSCREEN_DEFAULT),
+    .archive = true
 };
 static cvar_t vid_fullscreen_mode = {
-    "vid_fullscreen_mode",
-    stringify(VID_MODE_FULLSCREEN_DEFAULT),
-    true
+    .name = "vid_fullscreen_mode",
+    .string = stringify(VID_MODE_FULLSCREEN_DEFAULT),
+    .archive = true
 };
 static cvar_t vid_windowed_mode = {
-    "vid_windowed_mode",
-    stringify(VID_MODE_WINDOWED),
-    true
+    .name = "vid_windowed_mode",
+    .string = stringify(VID_MODE_WINDOWED),
+    .archive = true
 };
 
 static cvar_t vid_wait = { "vid_wait", "0" };
