@@ -159,8 +159,8 @@ SV_HullForEntity(edict_t *ent, vec3_t mins, vec3_t maxs, vec3_t offset)
 // calculate an offset value to center the origin
 	VectorSubtract(hull->clip_mins, mins, offset);
 	VectorAdd(offset, ent->v.origin, offset);
-    } else {			// create a temp hull from bounding box sizes
-
+    } else {
+	/* create a temp hull from bounding box sizes */
 	VectorSubtract(ent->v.mins, maxs, hullmins);
 	VectorSubtract(ent->v.maxs, mins, hullmaxs);
 	hull = SV_HullForBox(hullmins, hullmaxs);
