@@ -459,7 +459,7 @@ CL_Disconnect(void)
 	    CL_Stop_f();
 
 	final[0] = clc_stringcmd;
-	strcpy(final + 1, "drop");
+	strcpy((char *)final + 1, "drop");
 	Netchan_Transmit(&cls.netchan, 6, final);
 	Netchan_Transmit(&cls.netchan, 6, final);
 	Netchan_Transmit(&cls.netchan, 6, final);
@@ -1278,6 +1278,10 @@ Host_WriteConfiguration(void)
 	Key_WriteBindings(f);
 	Cvar_WriteVariables(f);
 
+	/* Save the mlook state (rarely used as an actual key binding) */
+	if (in_mlook.state & 1)
+	    fprintf(f, "+mlook\n");
+
 	fclose(f);
     }
 }
@@ -1514,14 +1518,15 @@ Host_Init(quakeparms_t *parms)
     IN_Init();
 #endif
 
+    Hunk_AllocName(0, "-HOST_HUNKLEVEL-");
+    host_hunklevel = Hunk_LowMark();
+
     Cbuf_InsertText("exec quake.rc\n");
+    Cbuf_Execute();
     Cbuf_AddText
 	("echo Type connect <internet address> or use GameSpy to connect to a game.\n");
     Cbuf_AddText("cl_warncmd 1\n");
 
-    Hunk_AllocName(0, "-HOST_HUNKLEVEL-");
-    host_hunklevel = Hunk_LowMark();
-
     host_initialized = true;
 
     Con_Printf("\nClient Version %4.2f (Build %04d)\n\n", VERSION,
