@@ -26,6 +26,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "render.h"
 #include "screen.h"
 #include "sys.h"
+#include "vid.h"
 #include "view.h"
 
 entity_t r_worldentity;
@@ -86,13 +87,25 @@ cvar_t r_wateralpha = { "r_wateralpha", "1" };
 cvar_t r_dynamic = { "r_dynamic", "1" };
 cvar_t r_novis = { "r_novis", "0" };
 cvar_t r_netgraph = { "r_netgraph", "0" };
+cvar_t r_waterwarp = { "r_waterwarp", "1" };
 
 cvar_t r_fullbright = {
     .name = "r_fullbright",
     .string = "0",
-    .readonly = true
+    .flags = CVAR_DEVELOPER
+};
+cvar_t gl_keeptjunctions = {
+    .name = "gl_keeptjunctions",
+    .string = "1",
+    .flags = CVAR_OBSOLETE
+};
+cvar_t gl_reporttjunctions = {
+    .name = "gl_reporttjunctions",
+    .string = "0",
+    .flags = CVAR_OBSOLETE
 };
 
+cvar_t gl_finish = { "gl_finish", "0" };
 cvar_t gl_clear = { "gl_clear", "0" };
 cvar_t gl_cull = { "gl_cull", "1" };
 cvar_t gl_texsort = { "gl_texsort", "1" };
@@ -102,15 +115,16 @@ cvar_t gl_polyblend = { "gl_polyblend", "1" };
 cvar_t gl_flashblend = { "gl_flashblend", "1" };
 cvar_t gl_playermip = { "gl_playermip", "0" };
 cvar_t gl_nocolors = { "gl_nocolors", "0" };
-cvar_t gl_keeptjunctions = { "gl_keeptjunctions", "1" };
-cvar_t gl_reporttjunctions = { "gl_reporttjunctions", "0" };
-cvar_t gl_finish = { "gl_finish", "0" };
+
+cvar_t _gl_lightmap_sort = { "_gl_lightmap_sort", "0" };
+cvar_t _gl_sky_mtex = { "_gl_sky_mtex", "1" };
+cvar_t _gl_allowgammafallback = { "_gl_allowgammafallback", "1" };
 
 /*
 =================
 R_CullBox
 
-Returns true if the box is completely outside the frustom
+Returns true if the box is completely outside the frustum
 =================
 */
 qboolean
@@ -183,7 +197,6 @@ R_GetSpriteFrame(entity_t *currententity)
 	    if (pintervals[i] > targettime)
 		break;
 	}
-
 	pspriteframe = pspritegroup->frames[i];
     }
 
@@ -223,7 +236,6 @@ R_DrawSpriteModel(entity_t *e)
     glColor3f(1, 1, 1);
 
     GL_DisableMultitexture();
-
     GL_Bind(frame->gl_texturenum);
 
     glEnable(GL_ALPHA_TEST);
@@ -262,7 +274,6 @@ R_DrawSpriteModel(entity_t *e)
 =============================================================
 */
 
-
 #define NUMVERTEXNORMALS 162
 
 float r_avertexnormals[NUMVERTEXNORMALS][3] = {
@@ -328,7 +339,6 @@ GL_DrawAliasFrame(aliashdr_t *paliashdr, int posenum)
     }
 }
 
-
 /*
 =============
 GL_DrawAliasShadow
@@ -427,7 +437,6 @@ R_SetupAliasFrame(int frame, aliashdr_t *paliashdr)
 /*
 =================
 R_DrawAliasModel
-
 =================
 */
 static void
@@ -451,14 +460,12 @@ R_DrawAliasModel(entity_t *e)
     if (R_CullBox(mins, maxs))
 	return;
 
-
     VectorCopy(currententity->origin, r_entorigin);
     VectorSubtract(r_origin, r_entorigin, modelorg);
 
     //
     // get lighting information
     //
-
     ambientlight = shadelight = R_LightPoint(currententity->origin);
 
     // allways give the gun some light
@@ -473,7 +480,7 @@ R_DrawAliasModel(entity_t *e)
 
 	    if (add > 0) {
 		ambientlight += add;
-		//ZOID models should be affected by dlights as well
+		/* ZOID: models should be affected by dlights as well */
 		shadelight += add;
 	    }
 	}
@@ -489,7 +496,6 @@ R_DrawAliasModel(entity_t *e)
     if (!strcmp(clmodel->name, "progs/player.mdl")) {
 	if (ambientlight < 8)
 	    ambientlight = shadelight = 8;
-
     } else if (!strcmp(clmodel->name, "progs/flame.mdl")
 	       || !strcmp(clmodel->name, "progs/flame2.mdl")) {
 	// HACK HACK HACK -- no fullbright colors, so make torches full light
@@ -517,9 +523,7 @@ R_DrawAliasModel(entity_t *e)
     //
     // draw all the triangles
     //
-
     GL_DisableMultitexture();
-
     glPushMatrix();
     R_RotateForEntity(e);
 
@@ -580,7 +584,6 @@ R_DrawAliasModel(entity_t *e)
 	glColor4f(1, 1, 1, 1);
 	glPopMatrix();
     }
-
 }
 
 //=============================================================================
@@ -627,9 +630,9 @@ R_MarkLeaves(void)
 }
 
 /*
-=============
+====================
 R_DrawEntitiesOnList
-=============
+====================
 */
 static void
 R_DrawEntitiesOnList(void)
@@ -737,6 +740,21 @@ R_DrawViewModel(void)
     glDepthRange(gldepthmin, gldepthmax);
 }
 
+/*
+ * GL_DrawBlendPoly
+ * - Render a polygon covering the whole screen
+ * - Used for full-screen color blending and approximated gamma correction
+ */
+static void
+GL_DrawBlendPoly(void)
+{
+    glBegin(GL_QUADS);
+    glVertex3f(10, 100, 100);
+    glVertex3f(10, -100, 100);
+    glVertex3f(10, -100, -100);
+    glVertex3f(10, 100, -100);
+    glEnd();
+}
 
 /*
 ============
@@ -746,36 +764,46 @@ R_PolyBlend
 static void
 R_PolyBlend(void)
 {
-    if (!gl_polyblend.value)
-	return;
-    if (!v_blend[3])
-	return;
+    float gamma = 1.0;
 
-    GL_DisableMultitexture();
+    if (!VID_IsFullScreen() || (!VID_SetGammaRamp &&
+				_gl_allowgammafallback.value)) {
+	gamma = v_gamma.value * v_gamma.value;
+	if (gamma < 0.25)
+	    gamma = 0.25;
+	else if (gamma > 1.0)
+	    gamma = 1.0;
+    }
 
-    glDisable(GL_ALPHA_TEST);
-    glEnable(GL_BLEND);
-    glDisable(GL_DEPTH_TEST);
-    glDisable(GL_TEXTURE_2D);
+    if ((gl_polyblend.value && v_blend[3]) || gamma < 1.0) {
+	GL_DisableMultitexture();
 
-    glLoadIdentity();
+	glDisable(GL_ALPHA_TEST);
+	glEnable(GL_BLEND);
+	glDisable(GL_DEPTH_TEST);
+	glDisable(GL_TEXTURE_2D);
 
-    glRotatef(-90, 1, 0, 0);	// put Z going up
-    glRotatef(90, 0, 0, 1);	// put Z going up
+	glLoadIdentity();
+	glRotatef(-90, 1, 0, 0);	// put Z going up
+	glRotatef(90, 0, 0, 1);		// put Z going up
 
-    glColor4fv(v_blend);
+	if (gl_polyblend.value && v_blend[3]) {
+	    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+	    glColor4fv(v_blend);
+	    GL_DrawBlendPoly();
+	}
+	if (gamma < 1.0) {
+	    glBlendFunc(GL_DST_COLOR, GL_ONE_MINUS_SRC_ALPHA);
+	    glColor4f(1, 1, 1, gamma);
+	    GL_DrawBlendPoly();
+	}
 
-    glBegin(GL_QUADS);
+	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 
-    glVertex3f(10, 100, 100);
-    glVertex3f(10, -100, 100);
-    glVertex3f(10, -100, -100);
-    glVertex3f(10, 100, -100);
-    glEnd();
-
-    glDisable(GL_BLEND);
-    glEnable(GL_TEXTURE_2D);
-    glEnable(GL_ALPHA_TEST);
+	glDisable(GL_BLEND);
+	glEnable(GL_TEXTURE_2D);
+	glEnable(GL_ALPHA_TEST);
+    }
 }
 
 
@@ -947,7 +975,7 @@ R_SetupGL(void)
     // FIXME - set depth dynamically for improved depth precision in smaller
     //         spaces
     //MYgluPerspective (r_refdef.fov_y, screenaspect, 4, 4096);
-    MYgluPerspective(r_refdef.fov_y, screenaspect, 4, 8192);
+    MYgluPerspective(r_refdef.fov_y, screenaspect, 4, 6144);
 
     if (mirror) {
 	if (mirror_plane->normal[2])
@@ -1024,7 +1052,7 @@ R_Clear(void)
 	gldepthmax = 0.5;
 	glDepthFunc(GL_LEQUAL);
     } else if (gl_ztrick.value) {
-	static int trickframe;
+	static unsigned int trickframe = 0;
 
 	if (gl_clear.value)
 	    glClear(GL_COLOR_BUFFER_BIT);
@@ -1141,8 +1169,10 @@ R_RenderView(void)
     if (!r_worldentity.model || !cl.worldmodel)
 	Sys_Error("%s: NULL worldmodel", __func__);
 
-    if (r_speeds.value) {
+    if (gl_finish.value || r_speeds.value)
 	glFinish();
+
+    if (r_speeds.value) {
 	time1 = Sys_DoubleTime();
 	c_brush_polys = 0;
 	c_alias_polys = 0;
@@ -1151,9 +1181,6 @@ R_RenderView(void)
 
     mirror = false;
 
-    if (gl_finish.value)
-	glFinish();
-
     R_Clear();
 
     // render normal view
