@@ -25,6 +25,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "pmove.h"
 #include "quakedef.h"
 #include "screen.h"
+#include "view.h"
 
 // FIXME - header hacks
 extern vrect_t scr_vrect;
@@ -275,7 +276,7 @@ byte gammatable[256];		// palette is sent through this
 
 
 #ifdef	GLQUAKE
-static byte ramps[3][256];
+static unsigned short ramps[3][256];
 float v_blend[4];		// rgba 0.0 - 1.0
 #endif // GLQUAKE
 
@@ -493,27 +494,20 @@ V_CalcBlend(void)
     float r, g, b, a, a2;
     int j;
 
-    r = 0;
-    g = 0;
-    b = 0;
-    a = 0;
+    r = g = b = a = 0;
 
-    for (j = 0; j < NUM_CSHIFTS; j++) {
-	if (!gl_cshiftpercent.value)
-	    continue;
-
-	a2 = ((cl.cshifts[j].percent * gl_cshiftpercent.value) / 100.0) /
-	    255.0;
-
-//              a2 = (cl.cshifts[j].percent/2)/255.0;
-	if (!a2)
-	    continue;
-	a = a + a2 * (1 - a);
-//Con_Printf ("j:%i a:%f\n", j, a);
-	a2 = a2 / a;
-	r = r * (1 - a2) + cl.cshifts[j].destcolor[0] * a2;
-	g = g * (1 - a2) + cl.cshifts[j].destcolor[1] * a2;
-	b = b * (1 - a2) + cl.cshifts[j].destcolor[2] * a2;
+    if (gl_cshiftpercent.value) {
+	for (j = 0; j < NUM_CSHIFTS; j++) {
+	    a2 = ((cl.cshifts[j].percent * gl_cshiftpercent.value) / 100.0) /
+		255.0;
+	    if (!a2)
+		continue;
+	    a = a + a2 * (1 - a);
+	    a2 = a2 / a;
+	    r = r * (1 - a2) + cl.cshifts[j].destcolor[0] * a2;
+	    g = g * (1 - a2) + cl.cshifts[j].destcolor[1] * a2;
+	    b = b * (1 - a2) + cl.cshifts[j].destcolor[2] * a2;
+	}
     }
 
     v_blend[0] = r / 255.0;
@@ -536,86 +530,33 @@ V_UpdatePalette
 void
 V_UpdatePalette(void)
 {
-    int i, j;
-    qboolean new;
-    byte *basepal, *newpal;
-    byte pal[768];
-    float r, g, b, a;
-    int ir, ig, ib;
+    int i;
     qboolean force;
 
     V_CalcPowerupCshift();
 
-    new = false;
-
-    for (i = 0; i < NUM_CSHIFTS; i++) {
-	if (cl.cshifts[i].percent != cl.prev_cshifts[i].percent) {
-	    new = true;
-	    cl.prev_cshifts[i].percent = cl.cshifts[i].percent;
-	}
-	for (j = 0; j < 3; j++)
-	    if (cl.cshifts[i].destcolor[j] != cl.prev_cshifts[i].destcolor[j]) {
-		new = true;
-		cl.prev_cshifts[i].destcolor[j] = cl.cshifts[i].destcolor[j];
-	    }
-    }
-
-// drop the damage value
+    /* drop the damage value */
     cl.cshifts[CSHIFT_DAMAGE].percent -= host_frametime * 150;
-    if (cl.cshifts[CSHIFT_DAMAGE].percent <= 0)
+    if (cl.cshifts[CSHIFT_DAMAGE].percent < 0)
 	cl.cshifts[CSHIFT_DAMAGE].percent = 0;
 
-// drop the bonus value
+    /* drop the bonus value */
     cl.cshifts[CSHIFT_BONUS].percent -= host_frametime * 100;
-    if (cl.cshifts[CSHIFT_BONUS].percent <= 0)
+    if (cl.cshifts[CSHIFT_BONUS].percent < 0)
 	cl.cshifts[CSHIFT_BONUS].percent = 0;
 
     force = V_CheckGamma();
-    if (!new && !force)
-	return;
 
-    V_CalcBlend();
-
-//Con_Printf("b: %4.2f %4.2f %4.2f %4.6f\n", v_blend[0],        v_blend[1],     v_blend[2],     v_blend[3]);
-
-    a = v_blend[3];
-    r = 255 * v_blend[0] * a;
-    g = 255 * v_blend[1] * a;
-    b = 255 * v_blend[2] * a;
-
-    a = 1 - a;
-    for (i = 0; i < 256; i++) {
-	ir = i * a + r;
-	ig = i * a + g;
-	ib = i * a + b;
-	if (ir > 255)
-	    ir = 255;
-	if (ig > 255)
-	    ig = 255;
-	if (ib > 255)
-	    ib = 255;
-
-	ramps[0][i] = gammatable[ir];
-	ramps[1][i] = gammatable[ig];
-	ramps[2][i] = gammatable[ib];
+    if (force) {
+	for (i = 0; i < 256; i++) {
+	    ramps[0][i] = gammatable[i] << 8;
+	    ramps[1][i] = gammatable[i] << 8;
+	    ramps[2][i] = gammatable[i] << 8;
+	}
+	if (VID_IsFullScreen() && VID_SetGammaRamp)
+	    VID_SetGammaRamp(ramps);
+	//VID_ShiftPalette(NULL);
     }
-
-    basepal = host_basepal;
-    newpal = pal;
-
-    for (i = 0; i < 256; i++) {
-	ir = basepal[0];
-	ig = basepal[1];
-	ib = basepal[2];
-	basepal += 3;
-
-	newpal[0] = ramps[0][ir];
-	newpal[1] = ramps[1][ig];
-	newpal[2] = ramps[2][ib];
-	newpal += 3;
-    }
-
-    VID_ShiftPalette(pal);
 }
 #else // !GLQUAKE
 /*
