@@ -884,6 +884,11 @@ Mod_LoadLeafs(lump_t *l)
     if (l->filelen % sizeof(*in))
 	Sys_Error("%s: funny lump size in %s", __func__, loadmodel->name);
     count = l->filelen / sizeof(*in);
+
+    /* FIXME - fail gracefully */
+    if (count > MAX_MAP_LEAFS)
+	Sys_Error("%s: model->numleafs > MAX_MAP_LEAFS\n", __func__);
+
     out = Hunk_AllocName(count * sizeof(*out), loadname);
 
     loadmodel->leafs = out;
@@ -988,7 +993,7 @@ Mod_LoadClipnodes(lump_t *l)
 =================
 Mod_MakeHull0
 
-Deplicate the drawing hull structure as a clipping hull
+Duplicate the drawing hull structure as a clipping hull
 =================
 */
 void
