@@ -34,17 +34,20 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  * - or just always warp underwater surfaces...
  * - issues with moving bsp models...
  */
-#define WATER_WARP_THROUGH_SURFACE
+//#define WATER_WARP_THROUGH_SURFACE
 
 #ifdef WATER_WARP_THROUGH_SURFACE
-#define WATER_WARP_TEST(surf)				\
-    ((((r_viewleaf->contents == CONTENTS_EMPTY)		\
-       && ((surf)->flags & SURF_UNDERWATER))		\
-      || ((r_viewleaf->contents != CONTENTS_EMPTY)	\
-       && !((surf)->flags & SURF_UNDERWATER)))		\
-     && !((surf)->flags & SURF_DONTWARP))
+/* FIXME - this version of the test is broken somehow... */
+#define WATER_WARP_TEST(surf)					\
+    (r_waterwarp.value && !((surf)->flags & SURF_DONTWARP) &&	\
+	((r_viewleaf->contents == CONTENTS_EMPTY &&		\
+	 ((surf)->flags & SURF_UNDERWATER))			\
+    ||								\
+	 (r_viewleaf->contents != CONTENTS_EMPTY &&		\
+         !((surf)->flags & SURF_UNDERWATER))))
 #else
-#define WATER_WARP_TEST(surf) ((surf)->flags & SURF_UNDERWATER)
+#define WATER_WARP_TEST(surf)		\
+    (r_waterwarp.value && ((surf)->flags & SURF_UNDERWATER))
 #endif
 
 int skytexturenum;
@@ -74,6 +77,22 @@ static int lm_used;
 // main memory so texsubimage can update properly
 static byte lightmaps[4 * MAX_LIGHTMAPS * BLOCK_WIDTH * BLOCK_HEIGHT];
 
+#if 0
+/*
+ * Bring the lightmap data under a common structure?
+ */
+struct lightmap_info {
+    glpoly_t *polys;
+    qboolean modified;
+    glRect_t rect;
+    int allocated[BLOCK_WIDTH];
+    byte data[4 * BLOCK_WIDTH * BLOCK_HEIGHT]; /* lightmaps */
+};
+
+static struct lightmap_info lightmaps[MAX_LIGHTMAPS];
+
+#endif
+
 // For gl_texsort 0
 static msurface_t *skychain = NULL;
 static msurface_t *waterchain = NULL;
@@ -81,10 +100,11 @@ static msurface_t *waterchain = NULL;
 static void R_RenderDynamicLightmaps(msurface_t *fa);
 
 /*
-==================
-R_AddDynamicLights
-==================
-*/
+ * ===================
+ * R_AddDynamicLights
+ * ===================
+ * Check all dynamic lights against this surface
+ */
 static void
 R_AddDynamicLights(msurface_t *surf)
 {
@@ -96,33 +116,32 @@ R_AddDynamicLights(msurface_t *surf)
     int i;
     int smax, tmax;
     mtexinfo_t *tex;
+    dlight_t *dl;
 
     smax = (surf->extents[0] >> 4) + 1;
     tmax = (surf->extents[1] >> 4) + 1;
     tex = surf->texinfo;
 
     for (lnum = 0; lnum < MAX_DLIGHTS; lnum++) {
-	if (!(surf->dlightbits & (1 << lnum)))
+	if (!(surf->dlightbits & (1U << lnum)))
 	    continue;		// not lit by this light
 
-	rad = cl_dlights[lnum].radius;
-	dist = DotProduct(cl_dlights[lnum].origin, surf->plane->normal) -
-	    surf->plane->dist;
+	dl = &cl_dlights[lnum];
+
+	rad = dl->radius;
+	dist = DotProduct(dl->origin, surf->plane->normal) - surf->plane->dist;
 	rad -= fabs(dist);
-	minlight = cl_dlights[lnum].minlight;
+	minlight = dl->minlight;
 	if (rad < minlight)
 	    continue;
 	minlight = rad - minlight;
 
-	for (i = 0; i < 3; i++) {
-	    impact[i] = cl_dlights[lnum].origin[i] -
-		surf->plane->normal[i] * dist;
-	}
+	for (i = 0; i < 3; i++)
+	    impact[i] = dl->origin[i] -	surf->plane->normal[i] * dist;
 
 	local[0] = DotProduct(impact, tex->vecs[0]) + tex->vecs[0][3];
-	local[1] = DotProduct(impact, tex->vecs[1]) + tex->vecs[1][3];
-
 	local[0] -= surf->texturemins[0];
+	local[1] = DotProduct(impact, tex->vecs[1]) + tex->vecs[1][3];
 	local[1] -= surf->texturemins[1];
 
 	for (t = 0; t < tmax; t++) {
@@ -144,12 +163,19 @@ R_AddDynamicLights(msurface_t *surf)
     }
 }
 
+/*
+ * Iterate through each lightmap for a surface
+ * msurface_t *s => the surface
+ * int n         => lightmap number
+ */
+#define for_each_surf_lightstyle(s, n) \
+	for ((n) = 0; (n) < MAXLIGHTMAPS && (s)->styles[n] != 255; (n)++)
 
 /*
 ===============
 R_BuildLightMap
 
-Combine and scale multiple lightmaps into the 8.8 format in blocklights
+Combine and scale multiple lightmaps into the correct format in blocklights
 ===============
 */
 static void
@@ -160,7 +186,7 @@ R_BuildLightMap(msurface_t *surf, byte *dest, int stride)
     int i, j, size;
     byte *lightmap;
     unsigned scale;
-    int maps;
+    int map;
     unsigned *bl;
 
     surf->cached_dlight = (surf->dlightframe == r_framecount);
@@ -181,15 +207,16 @@ R_BuildLightMap(msurface_t *surf, byte *dest, int stride)
 	blocklights[i] = 0;
 
 // add all the lightmaps
-    if (lightmap)
-	for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255;
-	     maps++) {
-	    scale = d_lightstylevalue[surf->styles[maps]];
-	    surf->cached_light[maps] = scale;	// 8.8 fraction
+    if (lightmap) {
+	for_each_surf_lightstyle(surf, map) {
+	    scale = d_lightstylevalue[surf->styles[map]];
+	    surf->cached_light[map] = scale;	// 8.8 fraction
 	    for (i = 0; i < size; i++)
 		blocklights[i] += lightmap[i] * scale;
 	    lightmap += size;	// skip to next lightmap
 	}
+    }
+
 // add all the dynamic lights
     if (surf->dlightframe == r_framecount)
 	R_AddDynamicLights(surf);
@@ -282,13 +309,20 @@ static qboolean mtexenabled = false;
 static GLenum oldtarget = GL_TEXTURE0_ARB;
 static int cnttextures[2] = { -1, -1 };	// cached
 
-static void
+/*
+ * Makes the given texture unit active
+ * FIXME: only aware of two texture units...
+ */
+void
 GL_SelectTexture(GLenum target)
 {
     if (!gl_mtexable || target == oldtarget)
 	return;
 
-    // FIXME - Understand this guff
+    /*
+     * Save the current texture unit's texture handle, select the new texture
+     * unit and update currenttexture
+     */
     qglActiveTextureARB(target);
     cnttextures[oldtarget - GL_TEXTURE0_ARB] = currenttexture;
     currenttexture = cnttextures[target - GL_TEXTURE0_ARB];
@@ -315,6 +349,68 @@ GL_EnableMultitexture(void)
     }
 }
 
+/*
+================
+DrawGLPoly
+================
+*/
+static void
+DrawGLPoly(glpoly_t *p)
+{
+    int i;
+    float *v;
+
+    glBegin(GL_POLYGON);
+    v = p->verts[0];
+    for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
+	glTexCoord2f(v[3], v[4]);
+	glVertex3fv(v);
+    }
+    glEnd();
+}
+
+static void
+DrawGLPolyLM(glpoly_t *p)
+{
+    int i;
+    float *v;
+
+    glBegin(GL_POLYGON);
+    v = p->verts[0];
+    for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
+	glTexCoord2f(v[5], v[6]);
+	glVertex3fv(v);
+    }
+    glEnd();
+}
+
+static void
+DrawGLPoly_2Ply(glpoly_t *p)
+{
+    int i;
+    float *v;
+
+    glBegin(GL_TRIANGLE_FAN);
+    v = p->verts[0];
+    for (i = 0; i < p->numverts; i++, v+= VERTEXSIZE) {
+	qglMultiTexCoord2fARB(GL_TEXTURE0_ARB, v[3], v[4]);
+	qglMultiTexCoord2fARB(GL_TEXTURE1_ARB, v[5], v[6]);
+	glVertex3fv(v);
+    }
+    glEnd();
+}
+
+static void
+WaterWarpCoord(vec3_t in, vec3_t out)
+{
+    /* Produce the warped coord */
+    out[0] = sin(in[1] * 0.05 + realtime) * sin(in[2] * 0.05 + realtime);
+    out[0] = out[0] * 8 + in[0];
+    out[1] = sin(in[0] * 0.05 + realtime) * sin(in[2] * 0.05 + realtime);
+    out[1] = out[1] * 8 + in[1];
+    out[2] = in[2];
+}
+
 /*
 ================
 DrawGLWaterPoly
@@ -335,15 +431,7 @@ DrawGLWaterPoly(glpoly_t *p)
     v = p->verts[0];
     for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
 	glTexCoord2f(v[3], v[4]);
-
-	nv[0] =
-	    v[0] + 8 * sin(v[1] * 0.05 + realtime) * sin(v[2] * 0.05 +
-							 realtime);
-	nv[1] =
-	    v[1] + 8 * sin(v[0] * 0.05 + realtime) * sin(v[2] * 0.05 +
-							 realtime);
-	nv[2] = v[2];
-
+	WaterWarpCoord(v, nv);
 	glVertex3fv(nv);
     }
     glEnd();
@@ -362,132 +450,115 @@ DrawGLWaterPolyLightmap(glpoly_t *p)
     v = p->verts[0];
     for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
 	glTexCoord2f(v[5], v[6]);
-
-	nv[0] =
-	    v[0] + 8 * sin(v[1] * 0.05 + realtime) * sin(v[2] * 0.05 +
-							 realtime);
-	nv[1] =
-	    v[1] + 8 * sin(v[0] * 0.05 + realtime) * sin(v[2] * 0.05 +
-							 realtime);
-	nv[2] = v[2];
-
+	WaterWarpCoord(v, nv);
 	glVertex3fv(nv);
     }
     glEnd();
 }
 
 
-#if 0
-/*
-================
-R_DrawSequentialPoly
-
-Systems that have fast state and texture changes can
-just do everything as it passes with no need to sort
-================
-*/
 static void
-R_DrawSequentialPoly(msurface_t *s)
+DrawGLWaterPoly_2Ply(glpoly_t *p)
 {
-    glpoly_t *p;
-    float *v;
     int i;
-    texture_t *t;
-
-    //
-    // normal lightmaped poly
-    //
-    if (!(s->flags & (SURF_DRAWSKY | SURF_DRAWTURB)) && !WATER_WARP_TEST(s)) {
-	p = s->polys;
-
-	t = R_TextureAnimation(s->texinfo->texture);
-	GL_Bind(t->gl_texturenum);
-	glBegin(GL_POLYGON);
-	v = p->verts[0];
-	for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
-	    glTexCoord2f(v[3], v[4]);
-	    glVertex3fv(v);
-	}
-	glEnd();
-
-	GL_Bind(lightmap_textures + s->lightmaptexturenum);
-	glEnable(GL_BLEND);
-	glBegin(GL_POLYGON);
-	v = p->verts[0];
-	for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
-	    glTexCoord2f(v[5], v[6]);
-	    glVertex3fv(v);
-	}
-	glEnd();
-
-	glDisable(GL_BLEND);
-
-	return;
-    }
-
-    //
-    // subdivided water surface warp
-    //
-    if (s->flags & SURF_DRAWTURB) {
-	GL_Bind(s->texinfo->texture->gl_texturenum);
-	EmitWaterPolys(s);
-	return;
-    }
-
-    //
-    // subdivided sky warp
-    //
-    if (s->flags & SURF_DRAWSKY) {
-	GL_Bind(solidskytexture);
-	speedscale = realtime * 8;
-	speedscale -= (int)speedscale;
-
-	EmitSkyPolys(s);
-
-	glEnable(GL_BLEND);
-	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-	GL_Bind(alphaskytexture);
-	speedscale = realtime * 16;
-	speedscale -= (int)speedscale;
-	EmitSkyPolys(s);
-	if (gl_lightmap_format == GL_LUMINANCE)
-	    glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
-
-	glDisable(GL_BLEND);
-    }
-
-    //
-    // underwater warped with lightmap
-    //
-    p = s->polys;
-
-    t = R_TextureAnimation(s->texinfo->texture);
-    GL_Bind(t->gl_texturenum);
-    DrawGLWaterPoly(p);
-
-    GL_Bind(lightmap_textures + s->lightmaptexturenum);
-    glEnable(GL_BLEND);
-    DrawGLWaterPolyLightmap(p);
-    glDisable(GL_BLEND);
-}
-#else
-/*
-================
-R_DrawSequentialPoly
-
-Systems that have fast state and texture changes can
-just do everything as it passes with no need to sort
-================
-*/
-static void
-R_DrawSequentialPoly(msurface_t *s)
-{
-    glpoly_t *p;
     float *v;
-    int i;
-    texture_t *t;
     vec3_t nv;
-    glRect_t *theRect;
+
+    glBegin(GL_TRIANGLE_FAN);
+    v = p->verts[0];
+    for (i = 0; i < p->numverts; i++, v+= VERTEXSIZE) {
+	qglMultiTexCoord2fARB(GL_TEXTURE0_ARB, v[3], v[4]);
+	qglMultiTexCoord2fARB(GL_TEXTURE1_ARB, v[5], v[6]);
+	WaterWarpCoord(v, nv);
+	glVertex3fv(nv);
+    }
+    glEnd();
+}
+
+static void
+DrawFlatGLPoly(glpoly_t *p)
+{
+    int i;
+    float *v;
+
+    srand((int)p);
+    glColor3f(rand() % 256 / 255.0, rand() % 256 / 255.0,
+	      rand() % 256 / 255.0);
+    glBegin(GL_POLYGON);
+    v = p->verts[0];
+    for (i = 0; i < p->numverts; i++, v += VERTEXSIZE)
+	glVertex3fv(v);
+    glEnd();
+}
+
+/*
+ * R_UploadLightmapUpdate
+ * Re-uploads the modified region of the given lightmap number
+ */
+static void
+R_UploadLightmapUpdate(int map)
+{
+    glRect_t *rect;
+    byte *pixels;
+    unsigned offset;
+
+    rect = &lightmap_rectchange[map];
+
+    offset = map * BLOCK_HEIGHT + rect->t; /* number of rows */
+    offset *= BLOCK_WIDTH;
+    offset += rect->l; /* x-offset */
+    offset *= lightmap_bytes;
+
+    pixels = lightmaps + offset;
+
+    /* set unpacking width to BLOCK_WIDTH, reset after */
+    glPixelStorei(GL_UNPACK_ROW_LENGTH, BLOCK_WIDTH);
+    glTexSubImage2D(GL_TEXTURE_2D,
+		    0,
+		    rect->l, /* x-offset */
+		    rect->t, /* y-offset */
+		    rect->w,
+		    rect->h,
+		    gl_lightmap_format,
+		    GL_UNSIGNED_BYTE,
+		    pixels);
+    glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
+
+    rect->l = BLOCK_WIDTH;
+    rect->t = BLOCK_HEIGHT;
+    rect->h = 0;
+    rect->w = 0;
+
+    c_lightmaps_uploaded++;
+}
+
+/*
+================
+R_DrawSequentialPoly
+
+Systems that have fast state and texture changes can
+just do everything as it passes with no need to sort
+================
+*/
+static void
+R_DrawSequentialPoly(msurface_t *s)
+{
+    glpoly_t *p;
+    int i;
+    texture_t *t;
+
+    /* A bit of a hack, but I don't like the layout of this function anyway */
+    if (r_drawflat.value) {
+	if (s->flags & (SURF_DRAWSKY | SURF_DRAWTURB)) {
+	    glpoly_t *p;
+	    for (p = s->polys; p; p = p->next)
+		DrawFlatGLPoly(p);
+	} else {
+	    DrawFlatGLPoly(s->polys);
+	}
+
+	return;
+    }
 
     //
     // normal lightmaped poly
@@ -507,53 +578,21 @@ R_DrawSequentialPoly(msurface_t *s)
 	    GL_Bind(lightmap_textures + s->lightmaptexturenum);
 	    i = s->lightmaptexturenum;
 	    if (lightmap_modified[i]) {
+		R_UploadLightmapUpdate(i);
 		lightmap_modified[i] = false;
-		theRect = &lightmap_rectchange[i];
-		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, theRect->t,
-				BLOCK_WIDTH, theRect->h, gl_lightmap_format,
-				GL_UNSIGNED_BYTE,
-				lightmaps + (i * BLOCK_HEIGHT +
-					     theRect->t) * BLOCK_WIDTH *
-				lightmap_bytes);
-		theRect->l = BLOCK_WIDTH;
-		theRect->t = BLOCK_HEIGHT;
-		theRect->h = 0;
-		theRect->w = 0;
-		c_lightmaps_uploaded++;
 	    }
 	    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
-	    glBegin(GL_POLYGON);
-	    v = p->verts[0];
-	    for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
-		qglMultiTexCoord2fARB(GL_TEXTURE0_ARB, v[3], v[4]);
-		qglMultiTexCoord2fARB(GL_TEXTURE1_ARB, v[5], v[6]);
-		glVertex3fv(v);
-	    }
-	    glEnd();
-	    return;
+	    DrawGLPoly_2Ply(p);
 	} else {
 	    p = s->polys;
 
 	    t = R_TextureAnimation(s->texinfo->texture);
 	    GL_Bind(t->gl_texturenum);
-	    glBegin(GL_POLYGON);
-	    v = p->verts[0];
-	    for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
-		glTexCoord2f(v[3], v[4]);
-		glVertex3fv(v);
-	    }
-	    glEnd();
+	    DrawGLPoly(p);
 
 	    GL_Bind(lightmap_textures + s->lightmaptexturenum);
 	    glEnable(GL_BLEND);
-	    glBegin(GL_POLYGON);
-	    v = p->verts[0];
-	    for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
-		glTexCoord2f(v[5], v[6]);
-		glVertex3fv(v);
-	    }
-	    glEnd();
-
+	    DrawGLPolyLM(p);
 	    glDisable(GL_BLEND);
 	}
 	return;
@@ -607,41 +646,15 @@ R_DrawSequentialPoly(msurface_t *s)
 	GL_Bind(lightmap_textures + s->lightmaptexturenum);
 	i = s->lightmaptexturenum;
 	if (lightmap_modified[i]) {
+	    R_UploadLightmapUpdate(i);
 	    lightmap_modified[i] = false;
-	    theRect = &lightmap_rectchange[i];
-	    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, theRect->t,
-			    BLOCK_WIDTH, theRect->h, gl_lightmap_format,
-			    GL_UNSIGNED_BYTE,
-			    lightmaps + (i * BLOCK_HEIGHT +
-					 theRect->t) * BLOCK_WIDTH *
-			    lightmap_bytes);
-	    theRect->l = BLOCK_WIDTH;
-	    theRect->t = BLOCK_HEIGHT;
-	    theRect->h = 0;
-	    theRect->w = 0;
-	    c_lightmaps_uploaded++;
 	}
 
 	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
-
-	glBegin(GL_TRIANGLE_FAN);
-	v = p->verts[0];
-	for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
-	    qglMultiTexCoord2fARB(GL_TEXTURE0_ARB, v[3], v[4]);
-	    qglMultiTexCoord2fARB(GL_TEXTURE1_ARB, v[5], v[6]);
-
-	    nv[0] =
-		v[0] + 8 * sin(v[1] * 0.05 + realtime) * sin(v[2] * 0.05 +
-							     realtime);
-	    nv[1] =
-		v[1] + 8 * sin(v[0] * 0.05 + realtime) * sin(v[2] * 0.05 +
-							     realtime);
-	    nv[2] = v[2];
-
-	    glVertex3fv(nv);
-	}
-	glEnd();
-
+	if (WATER_WARP_TEST(s))
+	    DrawGLWaterPoly_2Ply(p);
+	else
+	    DrawGLPoly_2Ply(p);
     } else {
 	p = s->polys;
 
@@ -655,28 +668,6 @@ R_DrawSequentialPoly(msurface_t *s)
 	glDisable(GL_BLEND);
     }
 }
-#endif
-
-
-/*
-================
-DrawGLPoly
-================
-*/
-static void
-DrawGLPoly(glpoly_t *p)
-{
-    int i;
-    float *v;
-
-    glBegin(GL_POLYGON);
-    v = p->verts[0];
-    for (i = 0; i < p->numverts; i++, v += VERTEXSIZE) {
-	glTexCoord2f(v[3], v[4]);
-	glVertex3fv(v);
-    }
-    glEnd();
-}
 
 
 /*
@@ -687,15 +678,15 @@ R_BlendLightmaps
 static void
 R_BlendLightmaps(void)
 {
-    int i, j;
+    int i;
     glpoly_t *p;
-    float *v;
-    glRect_t *theRect;
 
     if (r_fullbright.value)
 	return;
     if (!gl_texsort.value)
 	return;
+    if (r_drawflat.value)
+	return;
 
     glDepthMask(0);		// don't bother writing Z
 
@@ -717,32 +708,14 @@ R_BlendLightmaps(void)
 	    continue;
 	GL_Bind(lightmap_textures + i);
 	if (lightmap_modified[i]) {
+	    R_UploadLightmapUpdate(i);
 	    lightmap_modified[i] = false;
-	    theRect = &lightmap_rectchange[i];
-	    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, theRect->t,
-			    BLOCK_WIDTH, theRect->h, gl_lightmap_format,
-			    GL_UNSIGNED_BYTE,
-			    lightmaps + (i * BLOCK_HEIGHT +
-					 theRect->t) * BLOCK_WIDTH *
-			    lightmap_bytes);
-	    theRect->l = BLOCK_WIDTH;
-	    theRect->t = BLOCK_HEIGHT;
-	    theRect->h = 0;
-	    theRect->w = 0;
-	    c_lightmaps_uploaded++;
 	}
 	for (; p; p = p->chain) {
-	    if (WATER_WARP_TEST(p)) {
+	    if (WATER_WARP_TEST(p))
 		DrawGLWaterPolyLightmap(p);
-	    } else {
-		glBegin(GL_POLYGON);
-		v = p->verts[0];
-		for (j = 0; j < p->numverts; j++, v += VERTEXSIZE) {
-		    glTexCoord2f(v[5], v[6]);
-		    glVertex3fv(v);
-		}
-		glEnd();
-	    }
+	    else
+		DrawGLPolyLM(p);
 	}
     }
 
@@ -756,6 +729,65 @@ R_BlendLightmaps(void)
     glDepthMask(1);		// back to normal Z buffering
 }
 
+/*
+ * R_UpdateLightmapBlockRect
+ */
+static void
+R_UpdateLightmapBlockRect(msurface_t *fa)
+{
+    int map;
+    glRect_t *theRect;
+    byte *base;
+    int smax, tmax;
+
+    if (!r_dynamic.value)
+	return;
+
+    /* Check if any of this surface's lightmaps changed */
+    for_each_surf_lightstyle(fa, map)
+	if (d_lightstylevalue[fa->styles[map]] != fa->cached_light[map])
+	    goto dynamic;
+
+    /*
+     * 	fa->dlightframe == r_framecount	=> dynamic this frame
+     *  fa->cached_dlight		=> dynamic previously
+     */
+    if (fa->dlightframe == r_framecount || fa->cached_dlight) {
+    dynamic:
+	/*
+	 * Record that the lightmap block for this surface has been
+	 * modified. If necessary, increase the modified rectangle to include
+	 * this surface's allocatied sub-area.
+	 */
+	lightmap_modified[fa->lightmaptexturenum] = true;
+	theRect = &lightmap_rectchange[fa->lightmaptexturenum];
+	if (fa->light_t < theRect->t) {
+	    if (theRect->h)
+		theRect->h += theRect->t - fa->light_t;
+	    theRect->t = fa->light_t;
+	}
+	if (fa->light_s < theRect->l) {
+	    if (theRect->w)
+		theRect->w += theRect->l - fa->light_s;
+	    theRect->l = fa->light_s;
+	}
+	smax = (fa->extents[0] >> 4) + 1;
+	tmax = (fa->extents[1] >> 4) + 1;
+	if ((theRect->w + theRect->l) < (fa->light_s + smax))
+	    theRect->w = (fa->light_s - theRect->l) + smax;
+	if ((theRect->h + theRect->t) < (fa->light_t + tmax))
+	    theRect->h = (fa->light_t - theRect->t) + tmax;
+	base =
+	    lightmaps +
+	    fa->lightmaptexturenum * lightmap_bytes * BLOCK_WIDTH *
+	    BLOCK_HEIGHT;
+	base +=
+	    fa->light_t * BLOCK_WIDTH * lightmap_bytes +
+	    fa->light_s * lightmap_bytes;
+	R_BuildLightMap(fa, base, BLOCK_WIDTH * lightmap_bytes);
+    }
+}
+
 /*
 ================
 R_RenderBrushPoly
@@ -764,19 +796,20 @@ R_RenderBrushPoly
 void
 R_RenderBrushPoly(msurface_t *fa)
 {
+    int i;
     texture_t *t;
-    byte *base;
-    int maps;
-    glRect_t *theRect;
-    int smax, tmax;
 
     c_brush_polys++;
 
-    if (fa->flags & SURF_DRAWSKY) {	// warp texture, no lightmaps
+    if (fa->flags & SURF_DRAWSKY) {	// sky texture, no lightmaps
 	EmitBothSkyLayers(fa);
 	return;
     }
 
+    if (gl_mtexable) {
+	GL_SelectTexture(GL_TEXTURE0_ARB);
+	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
+    }
     t = R_TextureAnimation(fa->texinfo->texture);
     GL_Bind(t->gl_texturenum);
 
@@ -785,56 +818,33 @@ R_RenderBrushPoly(msurface_t *fa)
 	return;
     }
 
-    if (WATER_WARP_TEST(fa))
-	DrawGLWaterPoly(fa->polys);
-    else
-	DrawGLPoly(fa->polys);
-
-    // add the poly to the proper lightmap chain
-
-    fa->polys->chain = lightmap_polys[fa->lightmaptexturenum];
-    lightmap_polys[fa->lightmaptexturenum] = fa->polys;
-
-    // check for lightmap modification
-    for (maps = 0; maps < MAXLIGHTMAPS && fa->styles[maps] != 255; maps++)
-	if (d_lightstylevalue[fa->styles[maps]] != fa->cached_light[maps])
-	    goto dynamic;
-
-    if (fa->dlightframe == r_framecount	// dynamic this frame
-	|| fa->cached_dlight)	// dynamic previously
-    {
-      dynamic:
-	if (r_dynamic.value) {
-	    lightmap_modified[fa->lightmaptexturenum] = true;
-	    theRect = &lightmap_rectchange[fa->lightmaptexturenum];
-	    if (fa->light_t < theRect->t) {
-		if (theRect->h)
-		    theRect->h += theRect->t - fa->light_t;
-		theRect->t = fa->light_t;
-	    }
-	    if (fa->light_s < theRect->l) {
-		if (theRect->w)
-		    theRect->w += theRect->l - fa->light_s;
-		theRect->l = fa->light_s;
-	    }
-	    smax = (fa->extents[0] >> 4) + 1;
-	    tmax = (fa->extents[1] >> 4) + 1;
-	    if ((theRect->w + theRect->l) < (fa->light_s + smax))
-		theRect->w = (fa->light_s - theRect->l) + smax;
-	    if ((theRect->h + theRect->t) < (fa->light_t + tmax))
-		theRect->h = (fa->light_t - theRect->t) + tmax;
-	    base =
-		lightmaps +
-		fa->lightmaptexturenum * lightmap_bytes * BLOCK_WIDTH *
-		BLOCK_HEIGHT;
-	    base +=
-		fa->light_t * BLOCK_WIDTH * lightmap_bytes +
-		fa->light_s * lightmap_bytes;
-	    R_BuildLightMap(fa, base, BLOCK_WIDTH * lightmap_bytes);
+    if (gl_mtexable) {
+	/* All lightmaps are up to date... */
+	i = fa->lightmaptexturenum;
+	GL_EnableMultitexture();
+	GL_Bind(lightmap_textures + i);
+	if (lightmap_modified[i]) {
+	    R_UploadLightmapUpdate(i);
+	    lightmap_modified[i] = false;
 	}
+	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
+	DrawGLPoly_2Ply(fa->polys);
+	GL_DisableMultitexture();
+    } else {
+	if (WATER_WARP_TEST(fa))
+	    DrawGLWaterPoly(fa->polys);
+	else
+	    DrawGLPoly(fa->polys);
+
+	/* add the poly to the proper lightmap chain */
+	fa->polys->chain = lightmap_polys[fa->lightmaptexturenum];
+	lightmap_polys[fa->lightmaptexturenum] = fa->polys;
+
+	R_UpdateLightmapBlockRect(fa);
     }
 }
 
+
 /*
 ================
 R_RenderDynamicLightmaps
@@ -844,11 +854,6 @@ Multitexture
 static void
 R_RenderDynamicLightmaps(msurface_t *fa)
 {
-    byte *base;
-    int maps;
-    glRect_t *theRect;
-    int smax, tmax;
-
     c_brush_polys++;
 
     if (fa->flags & (SURF_DRAWSKY | SURF_DRAWTURB))
@@ -857,46 +862,7 @@ R_RenderDynamicLightmaps(msurface_t *fa)
     fa->polys->chain = lightmap_polys[fa->lightmaptexturenum];
     lightmap_polys[fa->lightmaptexturenum] = fa->polys;
 
-    // check for lightmap modification
-    for (maps = 0; maps < MAXLIGHTMAPS && fa->styles[maps] != 255; maps++)
-	if (d_lightstylevalue[fa->styles[maps]] != fa->cached_light[maps])
-	    goto dynamic;
-
-    /*
-     * 	fa->dlightframe == r_framecount	=> dynamic this frame
-     *  fa->cached_dlight		=> dynamic previously
-     */
-    if (fa->dlightframe == r_framecount || fa->cached_dlight) {
-    dynamic:
-	if (r_dynamic.value) {
-	    lightmap_modified[fa->lightmaptexturenum] = true;
-	    theRect = &lightmap_rectchange[fa->lightmaptexturenum];
-	    if (fa->light_t < theRect->t) {
-		if (theRect->h)
-		    theRect->h += theRect->t - fa->light_t;
-		theRect->t = fa->light_t;
-	    }
-	    if (fa->light_s < theRect->l) {
-		if (theRect->w)
-		    theRect->w += theRect->l - fa->light_s;
-		theRect->l = fa->light_s;
-	    }
-	    smax = (fa->extents[0] >> 4) + 1;
-	    tmax = (fa->extents[1] >> 4) + 1;
-	    if ((theRect->w + theRect->l) < (fa->light_s + smax))
-		theRect->w = (fa->light_s - theRect->l) + smax;
-	    if ((theRect->h + theRect->t) < (fa->light_t + tmax))
-		theRect->h = (fa->light_t - theRect->t) + tmax;
-	    base =
-		lightmaps +
-		fa->lightmaptexturenum * lightmap_bytes * BLOCK_WIDTH *
-		BLOCK_HEIGHT;
-	    base +=
-		fa->light_t * BLOCK_WIDTH * lightmap_bytes +
-		fa->light_s * lightmap_bytes;
-	    R_BuildLightMap(fa, base, BLOCK_WIDTH * lightmap_bytes);
-	}
-    }
+    R_UpdateLightmapBlockRect(fa);
 }
 
 /*
@@ -913,56 +879,6 @@ R_MirrorChain(msurface_t *s)
     mirror_plane = s->plane;
 }
 
-#if 0
-/*
-================
-R_DrawWaterSurfaces
-================
-*/
-void
-R_DrawWaterSurfaces(void)
-{
-    int i;
-    msurface_t *s;
-    texture_t *t;
-
-    if (r_wateralpha.value == 1.0)
-	return;
-
-    //
-    // go back to the world matrix
-    //
-    glLoadMatrixf(r_world_matrix);
-
-    glEnable(GL_BLEND);
-    glColor4f(1, 1, 1, r_wateralpha.value);
-    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
-
-    for (i = 0; i < cl.worldmodel->numtextures; i++) {
-	t = cl.worldmodel->textures[i];
-	if (!t)
-	    continue;
-	s = t->texturechain;
-	if (!s)
-	    continue;
-	if (!(s->flags & SURF_DRAWTURB))
-	    continue;
-
-	// set modulate mode explicitly
-	GL_Bind(t->gl_texturenum);
-
-	for (; s; s = s->texturechain)
-	    R_RenderBrushPoly(s);
-
-	t->texturechain = NULL;
-    }
-
-    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
-
-    glColor4f(1, 1, 1, 1);
-    glDisable(GL_BLEND);
-}
-#else
 /*
 ================
 R_DrawWaterSurfaces
@@ -1031,8 +947,6 @@ R_DrawWaterSurfaces(void)
     }
 }
 
-#endif
-
 /*
 ================
 DrawTextureChains
@@ -1056,6 +970,27 @@ DrawTextureChains(void)
 	return;
     }
 
+    if (gl_mtexable) {
+	/* Update all lightmaps */
+	for (i = 0; i < cl.worldmodel->numtextures; i++) {
+	    t = cl.worldmodel->textures[i];
+	    if (!t)
+		continue;
+	    s = t->texturechain;
+	    if (!s)
+		continue;
+	    if (i == skytexturenum)
+		continue;
+	    if (i == mirrortexturenum && !r_mirroralpha.value)
+		continue;
+	    if ((s->flags & SURF_DRAWTURB))
+		continue;
+
+	    for (; s; s = s->texturechain)
+		R_UpdateLightmapBlockRect(s);
+	}
+    }
+
     for (i = 0; i < cl.worldmodel->numtextures; i++) {
 	t = cl.worldmodel->textures[i];
 	if (!t)
@@ -1063,9 +998,9 @@ DrawTextureChains(void)
 	s = t->texturechain;
 	if (!s)
 	    continue;
-	if (i == skytexturenum)
+	if (i == skytexturenum) {
 	    R_DrawSkyChain(s);
-	else if (i == mirrortexturenum && r_mirroralpha.value != 1.0) {
+	} else if (i == mirrortexturenum && r_mirroralpha.value != 1.0) {
 	    R_MirrorChain(s);
 	    continue;
 	} else {
@@ -1074,9 +1009,53 @@ DrawTextureChains(void)
 	    for (; s; s = s->texturechain)
 		R_RenderBrushPoly(s);
 	}
+	t->texturechain = NULL;
+    }
+}
+
+static void
+DrawFlatTextureChains(void)
+{
+    int i;
+    msurface_t *s;
+    texture_t *t;
+
+    GL_DisableMultitexture();
+    glDisable(GL_TEXTURE_2D);
+
+    for (i = 0; i < cl.worldmodel->numtextures; i++) {
+
+	t = cl.worldmodel->textures[i];
+	if (!t)
+	    continue;
+
+	s = t->texturechain;
+	if (!s)
+	    continue;
+
+	/* sky and water polys are chained together! */
+	if (i == skytexturenum || (s->flags & SURF_DRAWTURB)) {
+	    for (; s; s = s->texturechain) {
+		glpoly_t *p;
+		for (p = s->polys; p; p = p->next)
+		    DrawFlatGLPoly(p);
+	    }
+	    t->texturechain = NULL;
+	    if (i == skytexturenum)
+		skychain = NULL;
+	    continue;
+	}
+
+	/* FIXME - handle mirror texture? */
+
+	for (; s; s = s->texturechain)
+	    DrawFlatGLPoly(s->polys);
 
 	t->texturechain = NULL;
     }
+
+    glEnable(GL_TEXTURE_2D);
+    glColor3f(1, 1, 1);
 }
 
 /*
@@ -1115,9 +1094,6 @@ R_DrawBrushModel(entity_t *e)
     if (R_CullBox(mins, maxs))
 	return;
 
-    glColor3f(1, 1, 1);
-    memset(lightmap_polys, 0, sizeof(lightmap_polys));
-
     VectorSubtract(r_refdef.vieworg, e->origin, modelorg);
     if (rotated) {
 	vec3_t temp;
@@ -1132,42 +1108,63 @@ R_DrawBrushModel(entity_t *e)
 
     psurf = &clmodel->surfaces[clmodel->firstmodelsurface];
 
-// calculate dynamic lighting for bmodel if it's not an
-// instanced model
-    if (clmodel->firstmodelsurface != 0 && !gl_flashblend.value) {
-	for (k = 0; k < MAX_DLIGHTS; k++) {
-	    if ((cl_dlights[k].die < cl.time) || (!cl_dlights[k].radius))
-		continue;
-
-	    R_MarkLights(&cl_dlights[k], 1 << k,
-			 clmodel->nodes + clmodel->hulls[0].firstclipnode);
-	}
-    }
-
     glPushMatrix();
     e->angles[0] = -e->angles[0];	// stupid quake bug
     R_RotateForEntity(e);
     e->angles[0] = -e->angles[0];	// stupid quake bug
 
-    //
-    // draw texture
-    //
-    for (i = 0; i < clmodel->nummodelsurfaces; i++, psurf++) {
-	// find which side of the node we are on
-	pplane = psurf->plane;
+    if (!r_drawflat.value) {
+	/*
+	 * calculate dynamic lighting for bmodel if it's not an instanced model
+	 */
+	memset(lightmap_polys, 0, sizeof(lightmap_polys));
 
+	if (clmodel->firstmodelsurface != 0 && !gl_flashblend.value) {
+	    for (k = 0; k < MAX_DLIGHTS; k++) {
+		if ((cl_dlights[k].die < cl.time) || (!cl_dlights[k].radius))
+		    continue;
+
+		R_MarkLights(&cl_dlights[k], 1 << k,
+			     clmodel->nodes + clmodel->hulls[0].firstclipnode);
+	    }
+	}
+    }
+
+    /*
+     * draw texture
+     */
+    if (r_drawflat.value)
+	glDisable(GL_TEXTURE_2D);
+    else
+	glColor3f(1, 1, 1);
+
+    for (i = 0; i < clmodel->nummodelsurfaces; i++, psurf++) {
+	/* find which side of the node we are on */
+	pplane = psurf->plane;
 	dot = DotProduct(modelorg, pplane->normal) - pplane->dist;
 
-	// draw the polygon
+	/* draw the polygon */
 	if (((psurf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) ||
 	    (!(psurf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON))) {
-	    if (gl_texsort.value)
+	    if (r_drawflat.value)
+		DrawFlatGLPoly(psurf->polys);
+	    else if (gl_texsort.value) {
+		/* FIXME - hack for dynamic lightmap updates... */
+		qboolean real_mtexable = gl_mtexable;
+		gl_mtexable = false;
 		R_RenderBrushPoly(psurf);
-	    else
+		gl_mtexable = real_mtexable;
+	    } else
 		R_DrawSequentialPoly(psurf);
 	}
     }
-    R_BlendLightmaps();
+
+    if (r_drawflat.value) {
+	glEnable(GL_TEXTURE_2D);
+	glColor3f(1, 1, 1);
+    } else {
+	R_BlendLightmaps();
+    }
 
     glPopMatrix();
 }
@@ -1228,74 +1225,67 @@ R_RecursiveWorldNode(mnode_t *node)
 
     switch (plane->type) {
     case PLANE_X:
-	dot = modelorg[0] - plane->dist;
-	break;
     case PLANE_Y:
-	dot = modelorg[1] - plane->dist;
-	break;
     case PLANE_Z:
-	dot = modelorg[2] - plane->dist;
+	dot = modelorg[plane->type - PLANE_X] - plane->dist;
 	break;
     default:
 	dot = DotProduct(modelorg, plane->normal) - plane->dist;
 	break;
     }
 
-    if (dot >= 0)
-	side = 0;
-    else
-	side = 1;
+    side = (dot >= 0) ? 0 : 1;
 
-// recurse down the children, front side first
+    /* recurse down the children, front side first */
     R_RecursiveWorldNode(node->children[side]);
 
-// draw stuff
+    /* draw stuff */
     c = node->numsurfaces;
 
     if (c) {
 	surf = cl.worldmodel->surfaces + node->firstsurface;
 
-	if (dot < 0 - BACKFACE_EPSILON)
+	if (dot < -BACKFACE_EPSILON)
 	    side = SURF_PLANEBACK;
 	else if (dot > BACKFACE_EPSILON)
 	    side = 0;
-	{
-	    for (; c; c--, surf++) {
-		if (surf->visframe != r_framecount)
-		    continue;
 
-		// don't backface underwater surfaces, because they warp
-		if (!WATER_WARP_TEST(surf)
-		    && ((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK)))
-		    continue;	// wrong side
+	for (; c; c--, surf++) {
+	    if (surf->visframe != r_framecount)
+		continue;
 
-		// if sorting by texture, just store it out
-		if (gl_texsort.value) {
-		    if (!mirror
-			|| surf->texinfo->texture !=
-			cl.worldmodel->textures[mirrortexturenum]) {
-			surf->texturechain =
-			    surf->texinfo->texture->texturechain;
-			surf->texinfo->texture->texturechain = surf;
-		    }
-		} else if (surf->flags & SURF_DRAWSKY) {
-		    surf->texturechain = skychain;
-		    skychain = surf;
-		} else if (surf->flags & SURF_DRAWTURB) {
-		    surf->texturechain = waterchain;
-		    waterchain = surf;
-		} else
-		    R_DrawSequentialPoly(surf);
-	    }
+	    // don't backface underwater surfaces, because they warp
+	    if (!WATER_WARP_TEST(surf)
+		&& ((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK)))
+		continue;	// wrong side
+
+	    // if sorting by texture, just store it out
+	    if (gl_texsort.value) {
+		if (!mirror
+		    || surf->texinfo->texture !=
+		    cl.worldmodel->textures[mirrortexturenum]) {
+		    surf->texturechain = surf->texinfo->texture->texturechain;
+		    surf->texinfo->texture->texturechain = surf;
+		}
+	    } else if (r_drawflat.value) {
+		/* don't bother with chains for flat polys... */
+		R_DrawSequentialPoly(surf);
+	    } else if (surf->flags & SURF_DRAWSKY) {
+		surf->texturechain = skychain;
+		skychain = surf;
+	    } else if (surf->flags & SURF_DRAWTURB) {
+		surf->texturechain = waterchain;
+		waterchain = surf;
+	    } else
+		R_DrawSequentialPoly(surf);
 	}
-
     }
-// recurse down the back side
+
+    /* recurse down the back side */
     R_RecursiveWorldNode(node->children[side ? 0 : 1]);
 }
 
 
-
 /*
 =============
 R_DrawWorld
@@ -1317,11 +1307,34 @@ R_DrawWorld(void)
     glColor3f(1, 1, 1);
     memset(lightmap_polys, 0, sizeof(lightmap_polys));
 
-    R_RecursiveWorldNode(cl.worldmodel->nodes);
+    if (r_drawflat.value || _gl_drawhull.value) {
+	GL_DisableMultitexture();
+	glDisable(GL_TEXTURE_2D);
+    }
 
-    DrawTextureChains();
+    if (_gl_drawhull.value) {
+	switch ((int)_gl_drawhull.value) {
+	case 1:
+	case 2:
+	    /* all preparation done when variable is set */
+	    R_DrawWorldHull();
+	    break;
+	default:
+	    /* FIXME: Error? should never happen... */
+	    break;
+	}
+	glEnable(GL_TEXTURE_2D);
+	glColor3f(1.0, 1.0, 1.0);
+    } else {
+	R_RecursiveWorldNode(cl.worldmodel->nodes);
 
-    R_BlendLightmaps();
+	if (r_drawflat.value) {
+	    DrawFlatTextureChains();
+	} else {
+	    DrawTextureChains();
+	    R_BlendLightmaps();
+	}
+    }
 }
 
 /*
@@ -1399,7 +1412,6 @@ AllocBlock(int w, int h, int *x, int *y)
 mvertex_t *r_pcurrentvertbase;
 
 static model_t *currentmodel;
-static int nColinElim;
 
 /*
 ================
@@ -1470,45 +1482,6 @@ BuildSurfaceDisplayList(msurface_t *fa)
 	poly->verts[i][5] = s;
 	poly->verts[i][6] = t;
     }
-
-    //
-    // remove co-linear points - Ed
-    //
-    if (!gl_keeptjunctions.value && !(fa->flags & SURF_UNDERWATER)) {
-	for (i = 0; i < lnumverts; ++i) {
-	    vec3_t v1, v2;
-	    float *prev, *this, *next;
-
-	    prev = poly->verts[(i + lnumverts - 1) % lnumverts];
-	    this = poly->verts[i];
-	    next = poly->verts[(i + 1) % lnumverts];
-
-	    VectorSubtract(this, prev, v1);
-	    VectorNormalize(v1);
-	    VectorSubtract(next, prev, v2);
-	    VectorNormalize(v2);
-
-	    // skip co-linear points
-#define COLINEAR_EPSILON 0.001
-	    if ((fabs(v1[0] - v2[0]) <= COLINEAR_EPSILON) &&
-		(fabs(v1[1] - v2[1]) <= COLINEAR_EPSILON) &&
-		(fabs(v1[2] - v2[2]) <= COLINEAR_EPSILON)) {
-		int j;
-
-		for (j = i + 1; j < lnumverts; ++j) {
-		    int k;
-
-		    for (k = 0; k < VERTEXSIZE; ++k)
-			poly->verts[j - 1][k] = poly->verts[j][k];
-		}
-		--lnumverts;
-		++nColinElim;
-		// retry next vertex next time, which is now current vertex
-		--i;
-	    }
-	}
-    }
-    poly->numverts = lnumverts;
 }
 
 /*
@@ -1593,24 +1566,80 @@ GL_BuildLightmaps(void)
 
     t1 = Sys_DoubleTime();
     cnt = 0;
-    for (j = 1; j < MAX_MODELS; j++) {
-	m = cl.model_precache[j];
-	if (!m)
-	    break;
-	if (m->name[0] == '*')
-	    continue;
-	r_pcurrentvertbase = m->vertexes;
-	currentmodel = m;
-	for (i = 0; i < m->numsurfaces; i++) {
-	    cnt++;
-	    GL_CreateSurfaceLightmap(m->surfaces + i);
-	    if (m->surfaces[i].flags & SURF_DRAWTURB)
+
+    if (_gl_lightmap_sort.value) {
+	/*
+	 * Create texture chains so that lightmaps get allocated in texture
+	 * order. That way, surfaces with the same textures are more likely to
+	 * be allocated in the same lightmap block and hence we will need less
+	 * state changes when rendering
+	 * FIXME - benchmark & pick one when texsort + multitexture is done).
+	 */
+	for (j = 1; j < MAX_MODELS; j++) {
+	    msurface_t *surf;
+	    texture_t *t;
+
+	    m = cl.model_precache[j];
+	    if (!m)
+		break;
+	    if (m->name[0] == '*')
 		continue;
-	    if (m->surfaces[i].flags & SURF_DRAWSKY)
+
+	    for (i = 0; i < m->numtextures; i++) {
+		t = m->textures[i];
+		if (t)
+		    t->texturechain = NULL;
+	    }
+
+	    for (i = 0; i < m->numsurfaces; i++) {
+		surf = m->surfaces + i;
+		surf->texturechain = surf->texinfo->texture->texturechain;
+		surf->texinfo->texture->texturechain = surf;
+	    }
+
+	    r_pcurrentvertbase = m->vertexes;
+	    currentmodel = m;
+
+	    for (i = 0; i < m->numtextures; i++) {
+		t = m->textures[i];
+		if (!t)
+		    continue;
+
+		for (surf = t->texturechain; surf; surf = surf->texturechain) {
+		    cnt++;
+		    GL_CreateSurfaceLightmap(surf);
+		    if (surf->flags & SURF_DRAWTURB)
+			continue;
+		    if (surf->flags & SURF_DRAWSKY)
+			continue;
+		    BuildSurfaceDisplayList(surf);
+		}
+		t->texturechain = NULL;
+	    }
+	}
+
+    } else {
+	for (j = 1; j < MAX_MODELS; j++) {
+	    m = cl.model_precache[j];
+	    if (!m)
+		break;
+
+	    if (m->name[0] == '*')
 		continue;
-	    BuildSurfaceDisplayList(m->surfaces + i);
+	    r_pcurrentvertbase = m->vertexes;
+	    currentmodel = m;
+	    for (i = 0; i < m->numsurfaces; i++) {
+		cnt++;
+		GL_CreateSurfaceLightmap(m->surfaces + i);
+		if (m->surfaces[i].flags & SURF_DRAWTURB)
+		    continue;
+		if (m->surfaces[i].flags & SURF_DRAWSKY)
+		    continue;
+		BuildSurfaceDisplayList(m->surfaces + i);
+	    }
 	}
     }
+
     t2 = Sys_DoubleTime();
     Con_DPrintf("Built lightmaps in %f seconds.(%i surfs).\n", t2 - t1, cnt);
     Con_DPrintf("AllocBlock time spent: %f seconds.\n", alloc_block_time);
