@@ -27,14 +27,8 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 #include <GL/glx.h>
 
-/* Compatibility with older GLX headers */
-#ifndef GLX_VERSION_1_4
-#define glXGetProcAddress glXGetProcAddressARB
-#endif
-
 #include <X11/keysym.h>
 #include <X11/cursorfont.h>
-
 #include <X11/extensions/xf86vmode.h>
 
 #include "common.h"
@@ -54,6 +48,10 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "x11_core.h"
 #include "in_x11.h"
 
+#if !defined(GLX_VERSION_1_4) && !defined(glXGetProcAddress)
+#define glXGetProcAddress glXGetProcAddressARB
+#endif
+
 #define WARP_WIDTH              320
 #define WARP_HEIGHT             200
 
@@ -72,10 +70,7 @@ static int old_mouse_x, old_mouse_y;
 
 static int scr_width, scr_height;
 
-static qboolean vidmode_ext = false;
-
-static XF86VidModeModeInfo **vidmodes;
-static int num_vidmodes;
+static XF86VidModeModeInfo saved_vidmode;
 static qboolean vidmode_active = false;
 
 /*-----------------------------------------------------------------------*/
@@ -101,6 +96,7 @@ const char *gl_extensions;
 
 static qboolean is8bit = false;
 qboolean gl_mtexable = false;
+static int gl_num_texture_units;
 
 // FIXME - lose this hack?
 qboolean isPermedia = false;
@@ -398,6 +394,10 @@ HandleEvents(void)
 		Key_Event(K_MOUSE3, true);
 	    else if (event.xbutton.button == 3)
 		Key_Event(K_MOUSE2, true);
+	    else if (event.xbutton.button == 4)
+		Key_Event(K_MWHEELUP, true);
+	    else if (event.xbutton.button == 5)
+		Key_Event(K_MWHEELDOWN, true);
 	    break;
 
 	case ButtonRelease:
@@ -407,6 +407,10 @@ HandleEvents(void)
 		Key_Event(K_MOUSE3, false);
 	    else if (event.xbutton.button == 3)
 		Key_Event(K_MOUSE2, false);
+	    else if (event.xbutton.button == 4)
+		Key_Event(K_MWHEELUP, false);
+	    else if (event.xbutton.button == 5)
+		Key_Event(K_MWHEELDOWN, false);
 	    break;
 
 	case CreateNotify:
@@ -425,29 +429,16 @@ HandleEvents(void)
 	IN_CenterMouse();
 }
 
-void
-VID_Shutdown(void)
-{
-    if (x_disp != NULL) {
-	if (ctx != NULL)
-	    glXDestroyContext(x_disp, ctx);
-	if (x_win != None)
-	    XDestroyWindow(x_disp, x_win);
-	if (vidmode_active)
-	    XF86VidModeSwitchToMode(x_disp, scrnum, vidmodes[0]);
-	XCloseDisplay(x_disp);
-    }
-    vidmode_active = false;
-    x_disp = NULL;
-    x_win = None;
-    ctx = NULL;
-}
+void (*VID_SetGammaRamp)(unsigned short ramp[3][256]);
+static unsigned short saved_gamma_ramp[3][256];
 
 void
 signal_handler(int sig)
 {
     printf("Received signal %d, exiting...\n", sig);
     XAutoRepeatOn(x_disp);
+    if (VID_SetGammaRamp)
+	VID_SetGammaRamp(saved_gamma_ramp);
     XCloseDisplay(x_disp);
     Sys_Quit();
     exit(0);
@@ -507,11 +498,12 @@ VID_SetPalette(unsigned char *palette)
 
     // FIXME - cache to disk like old QW client?
     for (i = 0; i < (1 << 15); i++) {
-	/* Maps
-	   000000000000000
-	   000000000011111 = Red  = 0x1F
-	   000001111100000 = Blue = 0x03E0
-	   111110000000000 = Grn  = 0x7C00
+	/*
+	 * Maps
+	 * 000000000000000
+	 * 000000000011111 = Red  = 0x1F
+	 * 000001111100000 = Blue = 0x03E0
+	 * 111110000000000 = Grn  = 0x7C00
 	 */
 	r = ((i & 0x1F) << 3) + 4;
 	g = ((i & 0x03E0) >> 2) + 4;
@@ -535,23 +527,58 @@ void
 CheckMultiTextureExtensions(void)
 {
     // FIXME - space at end of string? Check properly...
+    gl_mtexable = false;
     if (!COM_CheckParm("-nomtex")
 	&& strstr(gl_extensions, "GL_ARB_multitexture ")) {
-	Con_Printf("Found GL_ARB_multitexture\n");
+	Con_Printf("ARB multitexture extensions found.\n");
 
-	qglMultiTexCoord2fARB =
-	    (void *)glXGetProcAddress("glMultiTexCoord2fARB");
-	qglActiveTextureARB =
-	    (void *)glXGetProcAddress("glActiveTextureARB");
+	qglMultiTexCoord2fARB = glXGetProcAddress("glMultiTexCoord2fARB");
+	qglActiveTextureARB = glXGetProcAddress("glActiveTextureARB");
 
-	if (qglMultiTexCoord2fARB && qglActiveTextureARB) {
-	    Con_Printf("ARB multitexture extension enabled\n");
-	    gl_mtexable = true;
-	} else
+	/* Check how many texture units there actually are */
+	glGetIntegerv(GL_MAX_TEXTURE_UNITS, &gl_num_texture_units);
+
+	if (gl_num_texture_units < 2) {
+	    Con_Printf("Only %i texture units, multitexture disabled.\n",
+		       gl_num_texture_units);
+	} else if (!qglMultiTexCoord2fARB || !qglActiveTextureARB) {
 	    Con_Printf("ARB Multitexture symbols not found, disabled.\n");
+	} else {
+	    Con_Printf("ARB multitexture extension enabled\n"
+		       "-> %i texture units available\n",
+		       gl_num_texture_units);
+	    gl_mtexable = true;
+	}
     }
 }
 
+static void
+VID_SetXF86GammaRamp(unsigned short ramp[3][256])
+{
+    if (!x_disp)
+	Sys_Error("%s: x_disp == NULL!", __func__);
+
+    XF86VidModeSetGammaRamp(x_disp, scrnum, 256, ramp[0], ramp[1], ramp[2]);
+}
+
+/*
+ * Gamma_Init
+ * - Checks to see if gamma settings are available
+ * - Saves the current gamma settings
+ * - Sets the default gamma ramp function
+ */
+static void
+Gamma_Init()
+{
+    if (XF86VidModeGetGammaRamp(x_disp, scrnum, 256,
+				saved_gamma_ramp[0],
+				saved_gamma_ramp[1],
+				saved_gamma_ramp[2]))
+	VID_SetGammaRamp = VID_SetXF86GammaRamp;
+    else
+	VID_SetGammaRamp = NULL;
+}
+
 /*
 ===============
 GL_Init
@@ -639,8 +666,7 @@ VID_Init8bitPalette(void)
     int i;
 
     if (strstr(gl_extensions, "GL_EXT_shared_texture_palette") &&
-	(qglColorTableEXT =
-	 (void *)glXGetProcAddress("glColorTableEXT")) != NULL) {
+	(qglColorTableEXT = glXGetProcAddress("glColorTableEXT")) != NULL) {
 	char thePalette[256 * 3];
 	char *oldPalette, *newPalette;
 
@@ -657,25 +683,25 @@ VID_Init8bitPalette(void)
 	qglColorTableEXT(GL_SHARED_TEXTURE_PALETTE_EXT, GL_RGB, 256, GL_RGB,
 			 GL_UNSIGNED_BYTE, (void *)thePalette);
 	is8bit = true;
-    } else if (strstr(gl_extensions, "3DFX_set_global_palette") &&
-	       (qgl3DfxSetPaletteEXT =
-		(void *)glXGetProcAddress("gl3DfxSetPaletteEXT")) !=
-	       NULL) {
-	GLubyte table[256][4];
-	char *oldpal;
+    } else if (strstr(gl_extensions, "3DFX_set_global_palette")) {
+	qgl3DfxSetPaletteEXT = glXGetProcAddress("gl3DfxSetPaletteEXT");
+	if (qgl3DfxSetPaletteEXT) {
+	    GLubyte table[256][4];
+	    char *oldpal;
 
-	Con_SafePrintf("8-bit GL extensions (3dfx) enabled.\n");
-	oldpal = (char *)d_8to24table;	//d_8to24table3dfx;
-	for (i = 0; i < 256; i++) {
-	    table[i][2] = *oldpal++;
-	    table[i][1] = *oldpal++;
-	    table[i][0] = *oldpal++;
-	    table[i][3] = 255;
-	    oldpal++;
+	    Con_SafePrintf("8-bit GL extensions (3dfx) enabled.\n");
+	    oldpal = (char *)d_8to24table;	//d_8to24table3dfx;
+	    for (i = 0; i < 256; i++) {
+		table[i][2] = *oldpal++;
+		table[i][1] = *oldpal++;
+		table[i][0] = *oldpal++;
+		table[i][3] = 255;
+		oldpal++;
+	    }
+	    glEnable(GL_SHARED_TEXTURE_PALETTE_EXT);
+	    qgl3DfxSetPaletteEXT((GLuint *)table);
+	    is8bit = true;
 	}
-	glEnable(GL_SHARED_TEXTURE_PALETTE_EXT);
-	qgl3DfxSetPaletteEXT((GLuint *)table);
-	is8bit = true;
     }
 }
 
@@ -687,20 +713,20 @@ Check_Gamma (unsigned char *pal)
     unsigned char palette[768];
     int         i;
 
-    if ((i = COM_CheckParm ("-gamma")) == 0) {
-	if ((gl_renderer && strstr (gl_renderer, "Voodoo")) ||
-	    (gl_vendor && strstr (gl_vendor, "3Dfx"))) {
+    if ((i = COM_CheckParm("-gamma")) == 0) {
+	if ((gl_renderer && strstr(gl_renderer, "Voodoo")) ||
+	    (gl_vendor && strstr(gl_vendor, "3Dfx"))) {
 	    vid_gamma = 1;
 	} else {
 	    //vid_gamma = 0.7;	// default to 0.7 on non-3dfx hardware
 	    vid_gamma = 1.0;	// Be like the original...
 	}
     } else {
-	vid_gamma = Q_atof (com_argv[i + 1]);
+	vid_gamma = Q_atof(com_argv[i + 1]);
     }
 
     for (i = 0; i < 768; i++) {
-	f = pow ((pal[i] + 1) / 256.0, vid_gamma);
+	f = pow((pal[i] + 1) / 256.0, vid_gamma);
 	inf = f * 255 + 0.5;
 	if (inf < 0)
 	    inf = 0;
@@ -709,7 +735,7 @@ Check_Gamma (unsigned char *pal)
 	palette[i] = inf;
     }
 
-    memcpy (pal, palette, sizeof (palette));
+    memcpy(pal, palette, sizeof(palette));
 }
 */
 
@@ -720,6 +746,57 @@ VID_InitCvars(void)
     Cvar_RegisterVariable(&gl_ztrick);
 }
 
+/*
+ * Set the vidmode to the requested width/height if possible
+ * Return true if successful, false otherwise
+ */
+static qboolean
+VID_set_vidmode(int width, int height)
+{
+    int i, x, y, dist;
+    int best_dist = 9999999;
+    int num_modes;
+    XF86VidModeModeInfo **modes;
+    XF86VidModeModeInfo *mode = NULL;
+    qboolean mode_changed = false;
+
+    if (vidmode_active)
+	Sys_Error("%s: called while vidmode_active == true.", __func__);
+
+    XF86VidModeGetAllModeLines(x_disp, scrnum, &num_modes, &modes);
+    for (i = 0; i < num_modes; i++) {
+	if (width > modes[i]->hdisplay || height > modes[i]->vdisplay)
+	    continue;
+	x = width - modes[i]->hdisplay;
+	y = height - modes[i]->vdisplay;
+	dist = (x * x) + (y * y);
+	if (dist < best_dist) {
+	    best_dist = dist;
+	    mode = modes[i];
+	}
+    }
+
+    if (mode) {
+	mode_changed = XF86VidModeSwitchToMode(x_disp, scrnum, mode);
+	if (mode_changed) {
+	    vidmode_active = true;
+	    memcpy(&saved_vidmode, modes[0], sizeof(XF86VidModeModeInfo));
+	}
+    }
+
+    return mode_changed;
+}
+
+static void
+VID_restore_vidmode()
+{
+    if (vidmode_active) {
+	XF86VidModeSwitchToMode(x_disp, scrnum, &saved_vidmode);
+	if (saved_vidmode.privsize && saved_vidmode.private)
+	    XFree(saved_vidmode.private);
+    }
+}
+
 void
 VID_Init(unsigned char *palette)
 {
@@ -740,8 +817,8 @@ VID_Init(unsigned char *palette)
     Window root;
     XVisualInfo *visinfo;
     qboolean fullscreen = true;
+    qboolean vidmode_ext = false;
     int MajorVersion, MinorVersion;
-    int actualWidth, actualHeight;
 
 #ifdef QW_HACK
     S_Init();
@@ -757,7 +834,7 @@ VID_Init(unsigned char *palette)
 // interpret command-line params
 
 // set vid parameters
-    if ((i = COM_CheckParm("-window")) != 0)
+    if (COM_CheckParm("-window"))
 	fullscreen = false;
     if ((i = COM_CheckParm("-width")) != 0)
 	width = atoi(com_argv[i + 1]);
@@ -781,9 +858,13 @@ VID_Init(unsigned char *palette)
     if (vid.conheight < 200)
 	vid.conheight = 200;
 
-    if ((x_disp = XOpenDisplay(NULL)) == NULL) {
-	fprintf(stderr, "Error couldn't open the X display\n");
-	exit(1);
+    x_disp = XOpenDisplay(NULL);
+    if (!x_disp) {
+	if (getenv("DISPLAY"))
+	    Sys_Error("VID: Could not open display [%s]", getenv("DISPLAY"));
+	else
+	    Sys_Error("VID: Could not open local display\n");
+	exit(EXIT_FAILURE);
     }
 
     scrnum = DefaultScreen(x_disp);
@@ -791,10 +872,8 @@ VID_Init(unsigned char *palette)
 
     // Check video mode extension
     MajorVersion = MinorVersion = 0;
-    if (!XF86VidModeQueryVersion(x_disp, &MajorVersion, &MinorVersion)) {
-	vidmode_ext = false;
-    } else {
-	Con_Printf("Using XFree86-VidModeExtension Version %d.%d\n",
+    if (XF86VidModeQueryVersion(x_disp, &MajorVersion, &MinorVersion)) {
+	Con_Printf("Using XFree86-VidModeExtension Version %i.%i\n",
 		   MajorVersion, MinorVersion);
 	vidmode_ext = true;
     }
@@ -804,60 +883,30 @@ VID_Init(unsigned char *palette)
 	fprintf(stderr,
 		"qkHack: Error couldn't get an RGB, Double-buffered, "
 		"Depth visual\n");
-	exit(1);
+	exit(EXIT_FAILURE);
     }
 
-    if (vidmode_ext) {
-	int best_fit, best_dist, dist, x, y;
+    /* Attempt to set the vidmode if needed */
+    if (vidmode_ext && fullscreen)
+	fullscreen = VID_set_vidmode(width, height);
 
-	XF86VidModeGetAllModeLines(x_disp, scrnum, &num_vidmodes, &vidmodes);
-
-	// Are we going fullscreen?  If so, let's change video mode
-	if (fullscreen) {
-	    best_dist = 9999999;
-	    best_fit = -1;
-
-	    for (i = 0; i < num_vidmodes; i++) {
-		if (width > vidmodes[i]->hdisplay ||
-		    height > vidmodes[i]->vdisplay)
-		    continue;
-
-		x = width - vidmodes[i]->hdisplay;
-		y = height - vidmodes[i]->vdisplay;
-		dist = (x * x) + (y * y);
-		if (dist < best_dist) {
-		    best_dist = dist;
-		    best_fit = i;
-		}
-	    }
-
-	    if (best_fit != -1) {
-		actualWidth = vidmodes[best_fit]->hdisplay;
-		actualHeight = vidmodes[best_fit]->vdisplay;
-
-		// change to the mode
-		XF86VidModeSwitchToMode(x_disp, scrnum, vidmodes[best_fit]);
-		vidmode_active = true;
-	    } else
-		fullscreen = 0;
-	}
-    }
+    Gamma_Init();
 
     /* window attributes */
+    mask = CWBackPixel | CWColormap | CWEventMask;
     attr.background_pixel = 0;
-    attr.border_pixel = 0;
     attr.colormap = XCreateColormap(x_disp, root, visinfo->visual, AllocNone);
-
-    // FIXME - remove dependency (set masks later?)
     attr.event_mask = X_CORE_MASK | X_KEY_MASK | X_MOUSE_MASK;
+
     if (vidmode_active) {
-	mask = CWBackPixel | CWColormap | CWSaveUnder | CWBackingStore |
-	    CWEventMask | CWOverrideRedirect;
+	mask |= CWSaveUnder | CWBackingStore | CWOverrideRedirect;
 	attr.override_redirect = True;
 	attr.backing_store = NotUseful;
 	attr.save_under = False;
-    } else
-	mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;
+    } else {
+	mask |= CWBorderPixel;
+	attr.border_pixel = 0;
+    }
 
     x_win = XCreateWindow(x_disp, root, 0, 0, width, height,
 			  0, visinfo->depth, InputOutput,
@@ -913,6 +962,27 @@ VID_Init(unsigned char *palette)
     vid.recalc_refdef = 1;	// force a surface cache flush
 }
 
+void
+VID_Shutdown(void)
+{
+    if (VID_SetGammaRamp)
+	VID_SetGammaRamp(saved_gamma_ramp);
+
+    if (x_disp != NULL) {
+	if (ctx != NULL)
+	    glXDestroyContext(x_disp, ctx);
+	if (x_win != None)
+	    XDestroyWindow(x_disp, x_win);
+	if (vidmode_active)
+	    VID_restore_vidmode();
+	XCloseDisplay(x_disp);
+    }
+    vidmode_active = false;
+    x_disp = NULL;
+    x_win = None;
+    ctx = NULL;
+}
+
 void
 Sys_SendKeyEvents(void)
 {
@@ -990,7 +1060,14 @@ void
 VID_UnlockBuffer()
 {
 }
+
 void
 VID_LockBuffer()
 {
 }
+
+qboolean
+VID_IsFullScreen()
+{
+    return vidmode_active;
+}
