@@ -94,6 +94,8 @@ static const char *gl_vendor;
 static const char *gl_version;
 static const char *gl_extensions;
 
+static int gl_num_texture_units;
+
 qboolean DDActive;
 qboolean scr_skipupdate;
 
@@ -571,31 +573,37 @@ int texture_mode = GL_LINEAR;
 
 int texture_extension_number = 1;
 
-// FIXME - when is there no _WIN32 on windows NT?
-#ifdef _WIN32
 static void
 CheckMultiTextureExtensions(void)
 {
     // FIXME - Do proper substring testing (could be last extension, no space)
     // FIXME - Check for wglGetProcAddress errors...
-    if (strstr(gl_extensions, "GL_ARB_multitexture ")
-	&& !COM_CheckParm("-nomtex")) {
+    gl_mtexable = false;
+    if (!COM_CheckParm("-nomtex")
+	&& strstr(gl_extensions, "GL_ARB_multitexture ")) {
 	Con_Printf("ARB multitexture extensions found.\n");
+
 	qglMultiTexCoord2fARB =
 	    (void *)wglGetProcAddress("glMultiTexCoord2fARB");
-	qglActiveTextureARB = (void *)wglGetProcAddress("glActiveTextureARB");
-	gl_mtexable = true;
-    } else {
-	gl_mtexable = false;
+	qglActiveTextureARB =
+	    (void *)wglGetProcAddress("glActiveTextureARB");
+
+	/* Check how many texture units there actually are */
+	glGetIntegerv(GL_MAX_TEXTURE_UNITS, &gl_num_texture_units);
+
+	if (gl_num_texture_units < 2) {
+	    Con_Printf("Only %i texture units, multitexture disabled.\n",
+		       gl_num_texture_units);
+	} else if (!qglMultiTexCoord2fARB || !qglActiveTextureARB) {
+	    Con_Printf("ARB Multitexture symbols not found, disabled.\n");
+	} else {
+	    Con_Printf("ARB multitexture extension enabled\n"
+		       "-> %i texture units available\n",
+		       gl_num_texture_units);
+	    gl_mtexable = true;
+	}
     }
 }
-#else
-static void
-CheckMultiTextureExtensions(void)
-{
-    gl_mtexable = false;
-}
-#endif
 
 // FIXME - use or not?
 /* #define TYR__USE_VERTEX_ARRAY_EXT */
@@ -810,14 +818,33 @@ VID_SetPalette(unsigned char *palette)
     }
 }
 
-//static BOOL gammaworks;
+void (*VID_SetGammaRamp)(unsigned short ramp[3][256]);
+static unsigned short saved_gamma_ramp[3][256];
+
+static void
+VID_SetWinGammaRamp(unsigned short ramp[3][256])
+{
+    BOOL result;
+
+    result = SetDeviceGammaRamp(maindc, ramp);
+}
+
+void
+Gamma_Init(void)
+{
+    BOOL result = GetDeviceGammaRamp(maindc, saved_gamma_ramp);
+
+    if (result)
+	result = SetDeviceGammaRamp(maindc, saved_gamma_ramp);
+    if (result)
+	VID_SetGammaRamp = VID_SetWinGammaRamp;
+    else
+	VID_SetGammaRamp = NULL;
+}
 
 void
 VID_ShiftPalette(unsigned char *palette)
 {
-    // FIXME - header hacks
-    //extern byte ramps[3][256]; // commented out in VID_ShiftPalette
-
     //VID_SetPalette (palette);
     //gammaworks = SetDeviceGammaRamp (maindc, ramps);
 }
@@ -1797,6 +1824,7 @@ VID_Init(unsigned char *palette)
     VID_SetPalette(palette);
 
     VID_SetMode(vid_default, palette);
+    Gamma_Init();
 
     maindc = GetDC(mainwindow);
     bSetupPixelFormat(maindc);
@@ -1934,3 +1962,9 @@ VID_MenuKey(int key)
 	break;
     }
 }
+
+qboolean
+VID_IsFullScreen()
+{
+    return VID_GetModePtr(vid_modenum)->fullscreen;
+}
