@@ -375,7 +375,7 @@ SND_PickChannel(int entnum, int entchannel)
 	    continue;
 #endif
 	if (channel->end - paintedtime < life_left) {
-	    life_left = channels->end - paintedtime;
+	    life_left = channel->end - paintedtime;
 	    first_to_die = channel;
 	}
     }
