@@ -104,6 +104,9 @@ static int shiftmask_fl = 0;
 static long r_shift, g_shift, b_shift;
 static unsigned long r_mask, g_mask, b_mask;
 
+static XF86VidModeModeInfo saved_vidmode;
+static qboolean vidmode_active;
+
 static void
 shiftmask_init()
 {
@@ -189,7 +192,7 @@ xlib_rgb24(int r, int g, int b)
 }
 
 static void
-st2_fixup(XImage * framebuf, int x, int y, int width, int height)
+st2_fixup(XImage *framebuf, int x, int y, int width, int height)
 {
     int yi;
     unsigned char *src;
@@ -433,10 +436,63 @@ ResetSharedFrameBuffers(void)
 
 }
 
+
+/*
+ * Set the vidmode to the requested width/height if possible
+ * Return true if successful, false otherwise
+ */
+static qboolean
+VID_set_vidmode(int width, int height)
+{
+    int i, x, y, dist;
+    int best_dist = 9999999;
+    int num_modes;
+    XF86VidModeModeInfo **modes;
+    XF86VidModeModeInfo *mode = NULL;
+    qboolean mode_changed = false;
+
+    if (vidmode_active)
+	Sys_Error("%s: called while vidmode_active == true.", __func__);
+
+    XF86VidModeGetAllModeLines(x_disp, x_visinfo->screen, &num_modes, &modes);
+    for (i = 0; i < num_modes; i++) {
+	if (width > modes[i]->hdisplay || height > modes[i]->vdisplay)
+	    continue;
+	x = width - modes[i]->hdisplay;
+	y = height - modes[i]->vdisplay;
+	dist = (x * x) + (y * y);
+	if (dist < best_dist) {
+	    best_dist = dist;
+	    mode = modes[i];
+	}
+    }
+
+    if (mode) {
+	mode_changed = XF86VidModeSwitchToMode(x_disp, x_visinfo->screen,
+					       mode);
+	if (mode_changed) {
+	    vidmode_active = true;
+	    memcpy(&saved_vidmode, modes[0], sizeof(XF86VidModeModeInfo));
+	}
+    }
+
+    return mode_changed;
+}
+
+static void
+VID_restore_vidmode()
+{
+    if (vidmode_active) {
+	XF86VidModeSwitchToMode(x_disp, x_visinfo->screen, &saved_vidmode);
+	if (saved_vidmode.privsize && saved_vidmode.private)
+	    XFree(saved_vidmode.private);
+    }
+}
+
+
 // Called at startup to set up translation tables, takes 256 8 bit RGB values
 // the palette data will go away after the call, so it must be copied off if
 // the video driver will need it again
-
 void
 VID_Init(unsigned char *palette)
 {
@@ -444,6 +500,12 @@ VID_Init(unsigned char *palette)
     XVisualInfo template;
     int num_visuals;
     int template_mask;
+    XSetWindowAttributes attr;
+    unsigned long mask;
+    qboolean fullscreen = true;
+    qboolean vidmode_ext = false;
+    int MajorVersion, MinorVersion;
+    Window root;
 
 #ifdef QW_HACK
     S_Init();
@@ -465,16 +527,24 @@ VID_Init(unsigned char *palette)
     verbose = COM_CheckParm("-verbose");
 
 // open the display
-    x_disp = XOpenDisplay(0);
+    x_disp = XOpenDisplay(NULL);
     if (x_disp == NULL) {
 	if (getenv("DISPLAY"))
-	    Sys_Error("VID: Could not open display [%s]",
-		      getenv("DISPLAY"));
+	    Sys_Error("VID: Could not open display [%s]", getenv("DISPLAY"));
 	else
 	    Sys_Error("VID: Could not open local display");
+	exit(EXIT_FAILURE);
     }
-// catch signals so i can turn on auto-repeat
 
+    // Check video mode extension
+    MajorVersion = MinorVersion = 0;
+    if (XF86VidModeQueryVersion(x_disp, &MajorVersion, &MinorVersion)) {
+	Con_Printf("Using XFree86-VidModeExtension Version %i.%i\n",
+		   MajorVersion, MinorVersion);
+	vidmode_ext = true;
+    }
+
+    // catch signals so i can turn on auto-repeat
     {
 	struct sigaction sa;
 
@@ -489,6 +559,9 @@ VID_Init(unsigned char *palette)
 // for debugging only
     XSynchronize(x_disp, True);
 
+    if (COM_CheckParm("-window"))
+	fullscreen = false;
+
 // check for command-line window size
     if ((pnum = COM_CheckParm("-winsize"))) {
 	if (pnum >= com_argc - 2)
@@ -526,7 +599,7 @@ VID_Init(unsigned char *palette)
     else {
 	int screen;
 
-	screen = XDefaultScreen(x_disp);
+	screen = DefaultScreen(x_disp);
 	template.visualid =
 	    XVisualIDFromVisual(XDefaultVisual(x_disp, screen));
 	template_mask = VisualIDMask;
@@ -558,32 +631,38 @@ VID_Init(unsigned char *palette)
     }
 
     x_vis = x_visinfo->visual;
+    root = XRootWindow(x_disp, x_visinfo->screen);
 
-// setup attributes for main window
-    {
-	int attribmask = CWEventMask | CWColormap | CWBorderPixel;
-	XSetWindowAttributes attribs;
-	Colormap tmpcmap;
+    /* Attempt to set the vidmode if needed */
+    if (vidmode_ext && fullscreen)
+	fullscreen = VID_set_vidmode(vid.width, vid.height);
 
-	tmpcmap = XCreateColormap(x_disp, XRootWindow(x_disp,
-						      x_visinfo->screen),
-				  x_vis, AllocNone);
+    /* window attributes */
+    mask = CWEventMask | CWColormap | CWBackPixel;
+    attr.background_pixel = 0;
+    attr.colormap = XCreateColormap(x_disp, root, x_vis, AllocNone);
+    attr.event_mask = X_CORE_MASK | X_KEY_MASK | X_MOUSE_MASK;
 
-	// FIXME - remove dependency (set masks later?)
-	attribs.event_mask = X_CORE_MASK | X_KEY_MASK | X_MOUSE_MASK;
-	attribs.border_pixel = 0;
-	attribs.colormap = tmpcmap;
+    if (vidmode_active) {
+	mask |= CWSaveUnder | CWBackingStore | CWOverrideRedirect;
+	attr.override_redirect = True;
+	attr.backing_store = NotUseful;
+	attr.save_under = False;
+    } else {
+	mask |= CWBorderPixel;
+	attr.border_pixel = 0;
+    }
 
 // create the main window
-	x_win = XCreateWindow(x_disp, XRootWindow(x_disp, x_visinfo->screen), 0, 0,	// x, y
-			      vid.width, vid.height, 0,	// borderwidth
-			      x_visinfo->depth,
-			      InputOutput, x_vis, attribmask, &attribs);
-	XStoreName(x_disp, x_win, "xquake");
+    x_win = XCreateWindow(x_disp, XRootWindow(x_disp, x_visinfo->screen),
+			  0, 0,	// x, y
+			  vid.width, vid.height, 0,	// borderwidth
+			  x_visinfo->depth,
+			  InputOutput, x_vis, mask, &attr);
+    XStoreName(x_disp, x_win, "xquake");
 
-	if (x_visinfo->class != TrueColor)
-	    XFreeColormap(x_disp, tmpcmap);
-    }
+    //if (x_visinfo->class != TrueColor)
+    //XFreeColormap(x_disp, tmpcmap);
 
     if (x_visinfo->depth == 8) {
 	// create and upload the palette
@@ -593,6 +672,7 @@ VID_Init(unsigned char *palette)
 	    XSetWindowColormap(x_disp, x_win, x_cmap);
 	}
     }
+
 // create the GC
     {
 	XGCValues xgcvalues;
@@ -605,6 +685,17 @@ VID_Init(unsigned char *palette)
 // map the window
     XMapWindow(x_disp, x_win);
 
+    if (vidmode_active) {
+	XMoveWindow(x_disp, x_win, 0, 0);
+	XRaiseWindow(x_disp, x_win);
+
+	// FIXME - mouse may not be active...
+	IN_CenterMouse();
+	XFlush(x_disp);
+
+	XF86VidModeSetViewPort(x_disp, x_visinfo->screen, 0, 0);
+    }
+
 // wait for first exposure event
     {
 	XEvent event;
@@ -651,6 +742,14 @@ VID_Init(unsigned char *palette)
     vid.conheight = vid.height;
     vid.aspect = ((float)vid.height / (float)vid.width) * (320.0 / 240.0);
 
+#if 0
+    /*
+     * FIXME - make this configurable; e.g. this makes the aspect ratio look
+     *         correct on my non-standard (widescreen) laptop screen
+     */
+    vid.aspect = ((float)vid.height / (float)vid.width) * (1280.0 / 600.0);
+#endif
+
 //      XSynchronize(x_disp, False);
 }
 
@@ -699,6 +798,7 @@ void
 VID_Shutdown(void)
 {
     Con_Printf("VID_Shutdown\n");
+    VID_restore_vidmode();
     XAutoRepeatOn(x_disp);
     XCloseDisplay(x_disp);
 }
@@ -1082,7 +1182,8 @@ VID_Update(vrect_t *rects)
 
 }
 
-/* FIXME - Dither functions not used?
+#if 0
+/* FIXME - Dither functions not used? */
 static int  dither;
 
 static void
@@ -1102,9 +1203,8 @@ VID_DitherOff (void)
 	dither = 0;
     }
 }
-*/
 
-/* FIXME - some unused Sys_ functions?
+/* FIXME - some unused Sys_ functions? */
 static int
 Sys_OpenWindow (void)
 {
@@ -1125,15 +1225,15 @@ static void
 Sys_DisplayWindow (int window)
 {
 }
-*/
+#endif
 
 void
 Sys_SendKeyEvents(void)
 {
 }
 
-// FIXME - ever going to need this?
 #if 0
+/* FIXME - ever going to need this? */
 char *
 Sys_ConsoleInput(void)
 {
@@ -1153,13 +1253,12 @@ Sys_ConsoleInput(void)
     if (ready > 0) {
 	len = read(0, text, sizeof(text));
 	if (len >= 1) {
-	    text[len - 1] = 0;	// rip off the /n and terminate
+	    text[len - 1] = 0;	/* rip off the /n and terminate */
 	    return text;
 	}
     }
 
     return 0;
-
 }
 #endif
 
@@ -1182,15 +1281,15 @@ IN_Commands(void)
 	return;
 
     // FIXME - Need this consistant, robust
-    // FIXME - When fullscreen, don't release mouse
 
     // If we have the mouse, but are not in the game...
-    if (mouse_grab_active && key_dest != key_game) {
+    if (mouse_grab_active && key_dest != key_game && !vidmode_active) {
 	IN_UngrabMouse();
 	IN_UngrabKeyboard();
     }
     // If we don't have the mouse, but we're in the game and we want it...
-    if (!mouse_grab_active && key_dest == key_game && _windowed_mouse.value) {
+    if (!mouse_grab_active && key_dest == key_game &&
+	(_windowed_mouse.value || vidmode_active)) {
 	IN_GrabKeyboard();
 	IN_GrabMouse();
 	IN_CenterMouse();
@@ -1252,3 +1351,9 @@ void
 VID_UnlockBuffer(void)
 {
 }
+
+qboolean
+VID_IsFullScreen()
+{
+    return vidmode_active;
+}
